#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

    Airfoil and operations on it 

"""
import os
import ast
from datetime               import datetime, timedelta
from copy                   import copy
from typing                 import Type, override
from enum                   import StrEnum
from pathlib                import Path

import numpy as np

from base.math_util         import * 
from base.common_utils      import * 
from model.airfoil_geometry import Geometry_Splined, Geometry, Geometry_Bezier, Geometry_HicksHenne
from model.airfoil_geometry import Line, Side_Airfoil_Bezier

from model.xo2_driver       import Worker

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.WARNING)


#-------------------------------------------------------------------------------
# enums   
#-------------------------------------------------------------------------------

class usedAs (StrEnum):
    """ airfoil types for some usage semantics in application """
    NORMAL      = ""
    SEED        = "Seed"
    SEED_DESIGN = "Seed of design"
    REF         = "Reference" 
    DESIGN      = "Design"
    TARGET      = "Target"
    SECOND      = "Airfoil 2"
    FINAL       = "Final"

# geometry specification 

GEO_BASIC  = Geometry
GEO_SPLINE = Geometry_Splined




#--------------------------------------------------------------------------

class Airfoil:
    """ 

    Airfoil object to handle a airfoil direct related things  

    """
    isBlendAirfoil      = False
    isEdited            = False
    isExample           = False                      # vs. Example_Airfoil 
    isBezierBased       = False
    isHicksHenneBased   = False

    Extension           = '.dat'

    def __init__(self, x= None, y = None, name = None,
                 geometry : Type[Geometry]  = None, 
                 pathFileName = None,  workingDir= None):
        """
        Main constructor for new Airfoil

        Args:
            pathFileName: optional - string of existinng airfoil path and name 
            name: optional         - name of airfoil - no checks performed 
            x,y: optional          - the coordinates of airfoil 
            geometry: optional     - the geometry staretegy either GEO_BASIC, GEO_SPLNE...
            workingDir: optional   - base directoty where pathFileName is relative 
        """

        self._pathFileName = None
        if workingDir is not None:
            self._workingDir = os.path.normpath (workingDir)
        else: 
            self._workingDir = ''
        self._name          = name if name is not None else ''

        if not x is None: x = x if isinstance(x,np.ndarray) else np.asarray (x)
        self._x     = x
        if not y is None: y = y if isinstance(y,np.ndarray) else np.asarray (y)
        self._y     = y  

        self._isModified     = False
        self._isEdited       = False 
        self._isBlendAirfoil = False                        # is self blended from two other airfoils 

        if geometry is None: 
            self._geometry_class  = GEO_SPLINE              # geometry startegy 
        else:
            self._geometry_class  = geometry                # geometry startegy 
        self._geo            = None                         # selfs instance of geometry

        self._nPanelsNew     = None                         # repanel: no of panels - init via default
        self._le_bunch       = None                         # repanel: panel bunch at leading edge
        self._te_bunch       = None   	                    # repanel: panel bunch at trailing edge

        self._polarSet       = None                         # polarSet which is defined from outside 

        self._usedAs         = usedAs.NORMAL                # usage type of airfoil used by app <- AIRFOIL_TYPES
        self._propertyDict   = {}                           # multi purpose extra properties for an Airfoil
        self._file_datetime  = None                         # modification datetime of file 

        self._flap_setter    = None                         # proxy controller to flap self using Worker

        # pathFileName must exist if no coordinates were given 

        if (pathFileName is not None) and  (x is None or y is None): 
            pathFileName = os.path.normpath(pathFileName)   # ensure correct format
            if os.path.isabs (pathFileName):
                checkPath = pathFileName
            else:
                checkPath = os.path.join(self.workingDir, pathFileName)
            if not os.path.isfile(checkPath):
                self._name = "-- Error --"
                raise ValueError ("Airfoil file '%s' does not exist. Couldn't create Airfoil" % checkPath)
            else:
                self._pathFileName = pathFileName
                self._name = self.fileName_stem             # load will get the real name

        elif (pathFileName is not None) : 
                self._pathFileName = pathFileName

        elif (not name):
            self._name = "-- ? --"

        self._name_org      = self._name                    # original name for modification label
        self._fileName_org  = None                          # will hold original fileName 


    @classmethod
    def onDict(cls, dataDict, workingDir = None, geometry : Type[Geometry]  = None):
        """
        Alternate constructor for new Airfoil based on dictionary 

        Args:
            :dataDict: dictionary with "name" and "file" keys
            :workingDir: home of dictionary (paramter file) 
        """
        pathFileName  = fromDict(dataDict, "file", None)
        name          = fromDict(dataDict, "name", None)
        return cls(pathFileName = pathFileName, name = name, 
                   geometry = geometry, workingDir = workingDir)
        

    @classmethod
    def onFileType(cls, pathFileName, workingDir = None, geometry : Type[Geometry]  = None):
        """
        Alternate constructor for new Airfoil based on its file type retrieved from pathFileName

            '.dat'      - returns Airfoil 
            '.bez'      - returns Airfoil_Bezier 
            '.hicks'    - returns Airfoil_Hicks_Henne 

        Args:
            pathFileName: string of existinng airfoil path and name
            workingDir: optional working dir (if path is relative)
            geometry : geometry tyoe - only for .dat files 
        """

        ext = os.path.splitext(pathFileName)[1]

        if ext == Airfoil.Extension: 
            return Airfoil (pathFileName=pathFileName, workingDir=workingDir, geometry=geometry)

        elif ext == Airfoil_Bezier.Extension: 
            return Airfoil_Bezier (pathFileName=pathFileName, workingDir=workingDir)

        elif ext == Airfoil_Hicks_Henne.Extension: 
            return Airfoil_Hicks_Henne (pathFileName=pathFileName, workingDir=workingDir)
        else:
            raise ValueError (f"Unknown file extension '{ext}' for new airfoil")



    def _save (self, airfoilDict):
        """ stores the variables into the dataDict - returns the filled dict"""
        
        if self.isBlendAirfoil:
            toDict (airfoilDict, "name", self.name) 
        else:
            toDict (airfoilDict, "file", self.pathFileName) 
        return airfoilDict
 

    def __repr__(self) -> str:
        # overwritten to get a nice print string 
        info = f"'{self.fileName}'" if self.fileName else self.name
        return f"<{type(self).__name__} {info}>"


    def _handle_geo_changed (self, geo : Geometry = None):
        """ 
        callback from geometry when it was changed (by user) 
        
        Args:
            geo: optional alternative geometry which will be taken 
        """

        # load new coordinates from modified geometry 
        if geo is None: 
            self._x = self.geo.x
            self._y = self.geo.y
            modifications = self.geo.modification_dict
        else: 
            self._x = geo.x
            self._y = geo.y
            modifications = geo.modification_dict

        # set new name
        if not self._name_org: 
            self._name_org = self.name

        self.set_name (f"{self._name_org}_mods:{str(modifications)}")

        # if not DESIGN set new filename - take original filename if possible (without ...mod...)
        if not self.usedAsDesign:
            if self._fileName_org is None: 
                self._fileName_org = self.fileName

            fileName_stem = os.path.splitext(self._fileName_org)[0]
            fileName_ext  = os.path.splitext(self._fileName_org)[1]
            mod_string    = "_mod" if modifications else ""
            self.set_fileName (fileName_stem + mod_string + fileName_ext)

        self.set_isModified (True)
        logger.debug (f"{self} - geometry changed: {modifications} ")


    # ----------  Properties ---------------

    @property
    def x (self): return self._x

    @property
    def y (self): return self._y


    @property
    def geo (self) -> Geometry:
        """ the geometry of self"""
        if self._geo is None: 
            self._geo = self._geometry_class (self.x, self.y, onChange = self._handle_geo_changed)

        return self._geo
    

    def set_xy (self, x, y):
        """ set new coordinates """

        if not x is None: 
            x = x if isinstance(x,np.ndarray) else np.asarray (x)
            x = np.round(x,7)
        if not y is None: 
            y = y if isinstance(y,np.ndarray) else np.asarray (y)
            y = np.round(y,7)

        self._x     = x
        self._y     = y  
        self._geo    = None

        self.set_isModified (True)


    @property
    def workingDir (self) -> str: 
        """ working directory which is added to pathFileName"""
        return self._workingDir 
    
    def set_workingDir (self, aDir : str):
        self._workingDir = aDir


    @property
    def name (self) -> str: 
        """ name of airfoil"""

        return self._name 

    @property
    def name_to_show (self) -> str: 
        """ public name of airfoil - use fileName.stem - limited to 47 chars"""

        name = self.fileName_stem
        return name if len(name) <= 47 else f"{name[:47]}..."


    def set_name (self, newName, reset_original=False):
        """  Set name of the airfoil. 'reset_original' will also overwrite original name  
        Note:  This will not rename an existing airfoil (file)...
        """
        self._name = newName

        if not self._name_org or reset_original: 
            self._name_org = self.name

        self.set_isModified (True)

    @property
    def name_short (self):
        """ name of airfoil shortend at the beginning to 23 chars"""
        if len(self.name) <= 23:    return self.name
        else:                       return "..." + self.name[-20:]

    @property
    def info_as_html (self) -> str:
        """ comprehensive info about self as formatted html string"""

        used_as = f"{self.usedAs}: " if self.usedAs != usedAs.NORMAL else ""
        info = f"{used_as}{self.fileName}" 

        info = info + f"<br><br>in {self.pathName_abs}  " 

        t = f"<br>" + \
            f"<table>" + \
                f"<tr>" + \
                    f"<td>Thickness  </td>" + \
                    f"<td>{self.geo.max_thick:.2%}  </td>" + \
                    f"<td>at  </td>" + \
                    f"<td>{self.geo.max_thick_x:.2%}  </td>" + \
                f"</tr>" + \
                f"<tr>" + \
                    f"<td>Camber  </td>" + \
                    f"<td>{self.geo.max_camb:.2%}  </td>" + \
                    f"<td>at  </td>" + \
                    f"<td>{self.geo.max_camb_x:.2%}  </td>" + \
                f"</tr>" + \
                f"<tr>" + \
                f"</tr>" + \
                f"<tr>" + \
                    f"<td>Curvature LE  </td>" + \
                    f"<td>{self.geo.curvature.max_around_le:.0f}  </td>" + \
                    f"<td>TE    </td>" + \
                    f"<td>{self.geo.curvature.max_te:.0f}  </td>" + \
                f"</tr>" + \
            f"</table>"
        
        info = "<p style='white-space:pre'>" + info + t                     # no word wrap 
        return info 



    @property
    def polarSet (self):
        """ Property which is set from outside - Airfoil doesn't know about it... """ 
        return self._polarSet 
    def set_polarSet (self, aPolarSet):
        self._polarSet = aPolarSet


    @property
    def isEdited (self) -> bool: 
        """ True if airfoil is being edited, modified, ..."""
        return self._isModified
    def set_isEdited (self, aBool): 
        self.set_isModified (aBool) 


    @property
    def isModified (self) -> bool: 
        """ True if airfoil is being modifiied, ..."""
        return self._isModified
    def set_isModified (self, aBool): 
        self._isModified = aBool 

    @property
    def isExisting (self) -> bool:
        return not self.pathFileName is None


    @property
    def isLoaded (self) -> bool:
        return self._x is not None and len(self._x) > 10
    
    @property
    def isNormalized (self) -> bool:
        """ is LE at 0,0 and TE at 1,.. ?"""
        return self.geo.isNormalized
    
    @property
    def isBlendAirfoil (self) -> bool:
        """ is self blended out of two other airfoils"""
        return self._isBlendAirfoil
    def set_isBlendAirfoil (self, aBool): 
        self._isBlendAirfoil = aBool
    
    @property
    def nPanels (self) -> int: 
        """ number of panels """
        return self.geo.nPanels
      
    @property
    def nPoints (self) -> int: 
        """ number of coordinate points"""
        return self.geo.nPoints 

        
    @property
    def isSymmetrical(self) -> bool:
        """ true if max camber is 0.0 - so it's a symmetric airfoil"""
        return self.geo.isSymmetrical


    @property
    def isFlapped (self) -> bool:
        """ true if self is probably flapped"""
        return self.geo.isFlapped


    @property
    def isReflexed (self) -> bool:
        """ True if there is just one reversal on upper side"""
        return self.geo.curvature.isReflexed


    @property
    def isRearLoaded (self) -> bool:
        """ True if there is just one reversal on lower side"""
        return self.geo.curvature.isRearLoaded


    @property
    def isUpToDate (self) -> bool:
        """ true if the loaded airfoil is up to date with its file - none if no answer"""

        if os.path.isfile (self.pathFileName_abs) and self._file_datetime is not None:
            ts = os.path.getmtime(self.pathFileName_abs)                       # file modification timestamp of a file
            file_datetime = datetime.fromtimestamp(ts)                  # convert timestamp into DateTime object

            # add safety seconds (async stuff?) 
            if (self._file_datetime >= (file_datetime - timedelta(seconds=1))):
                return True
            else: 
                return False
        return None  


    @property
    def usedAs (self):
        """ usage type (enum usedAs) of self like DESIGN"""
        return self._usedAs
    def set_usedAs (self, aType): 
        if aType in usedAs:
            self._usedAs = aType

    @property
    def usedAsDesign (self): 
        """ short for self used as DESIGN """ 
        return self._usedAs == usedAs.DESIGN

    def useAsDesign (self, aBool=True): 
        """ set usedAs property to DESIGN"""
        if aBool: 
            self.set_usedAs (usedAs.DESIGN)
        else: 
            self.set_usedAs (usedAs.NORMAL)
            return

        # Design: get modifications dict from name

        if aBool: 
            try: 
                name_org, mods = self._name.split ("_mods:", 1)
                if mods and name_org:
                    self.geo._modification_dict = ast.literal_eval (mods)
                    self._name_org = name_org 
            except ValueError:
                self._name_org = self.name
                pass


    def get_property (self, name, default):
        """ returns free style property of self"""
        return fromDict (self._propertyDict, name, default )

    def set_property (self, name, aVal):
        """ set free style property of self"""
        return toDict (self._propertyDict, name, aVal )

    #-----------------------------------------------------------

    @property
    def pathFileName (self) -> str:
        """ path and filename of airfoil like './examples/JX-GT-15.dat' """
        return self._pathFileName


    def set_pathFileName (self, pathfileName : str, noCheck=False):
        """
        Set fullpaths of airfoils location and file  
        ! This will not move or copy the airfoil physically - use copyAs instead
        """
        if noCheck:
            self._pathFileName = pathfileName
        elif os.path.isfile(pathfileName):
            self._pathFileName = pathfileName
        elif self.workingDir and os.path.isfile(self.pathFileName_abs):
            self._pathFileName = pathfileName
        else:
            raise ValueError (f"Airfoil {pathfileName} does not exist. Couldn\'t be set")

    @property
    def pathFileName_abs (self) -> str:
        """ path including working dir and filename of airfoil like '/root/examples/JX-GT-15.dat' """

        if self.workingDir:
            pathFileName_abs =  os.path.join(self.workingDir, self.pathFileName)
        else: 
            pathFileName_abs =  self._pathFileName
        
        if not os.path.isabs (pathFileName_abs):
            pathFileName_abs = os.path.abspath(pathFileName_abs)       # will insert cwd 
        return pathFileName_abs


    def set_pathName (self, aDir : str, noCheck=False):
        """
        Set fullpaths of airfoils directory  
            ! This will not move or copy the airfoil physically
        """
        aDir_abs = os.path.join (self.workingDir, aDir)
        if noCheck or (os.path.isdir(aDir_abs)) or aDir == '':
            self._pathFileName = os.path.join (aDir, self.fileName)
        else:
            raise ValueError ("Directory \'%s\' does not exist. Couldn\'t be set" % aDir)

    @property
    def fileName (self):
        """ filename of airfoil like 'JX-GT-15.dat' """
        return os.path.basename(self.pathFileName) if self.pathFileName else ''
    
    @property
    def fileName_stem (self):
        """ stem of fileName like 'JX-GT-15' """
        return Path(self.fileName).stem if self.fileName else ''

    @property
    def fileName_ext (self):
        """ extension of fileName like '.dat'"""
        return os.path.splitext(self.fileName)[1] if self.fileName else ''


    def set_fileName (self, aFileName : str):
        """ set new fileName """
        if not aFileName: return 
        self._pathFileName = os.path.join (self.pathName, aFileName)

    def set_name_from_fileName (self):
        """ set current fileName as name of airfoil """
        self.set_name (self.fileName_stem)

    def set_fileName_from_name (self):
        """ set current fileName as name of airfoil """
        self.set_fileName (self.name + self.fileName_ext) 


    @property
    def pathName (self):
        """
        directory pathname of airfoil like '..\\myAirfoils\\'
        """
        if not self.pathFileName is None: 
            return os.path.dirname(self.pathFileName) 
        else:
            return ''

    @property
    def pathName_abs (self):
        """
        absolute directory pathname of airfoil like '\\root\\myAirfoils\\'
        """
        if not self.pathFileName_abs is None: 
            return os.path.dirname(self.pathFileName_abs)
        else:
            # fallback - current python working dir
            logger.warning (f"{self} has not pathFileName")
            return os.path.dirname(os.getcwd())


    def load (self, fromPath = None):
        """
        Loads airfoil coordinates from file. 
        pathFileName must be set before or fromPath must be defined.
        Load doesn't change self pathFileName
        """    

        if fromPath:
            sourcePathFile = fromPath
        else:
            sourcePathFile = self.pathFileName_abs

        if os.path.isfile (sourcePathFile):

            try:
                # read airfoil file into x,y 

                f = open(sourcePathFile, 'r')
                file_lines = f.readlines()
                f.close()
                self._name, self._x, self._y = self._loadLines(file_lines)

                self._geo = None                                            # reset geometry 

                # get modfication datetime of file 

                ts = os.path.getmtime(sourcePathFile)                       # file modification timestamp of a file
                self._file_datetime = datetime.fromtimestamp(ts)            # convert timestamp into DateTime object

            except ValueError as e:
                logger.error (f"{self} {e}")
                pass


    def _loadLines (self, file_lines : list[str]):
        """ extract name, x, y from file_lines"""
        # returns the name and x,y (np array) of the airfoil file 

        name = ''
        x = []
        y = []
        xvalPrev = -9999.9
        yvalPrev = -9999.9

        for i, line in enumerate(file_lines):
            if (i > 0): 
                splitline = line.strip().split()               # will remove all extra spaces
                if len(splitline) == 1:                        # couldn't split line - try tab as separator
                    splitline = line.strip().split("\t",1)
                if len(splitline) >= 2:                     
                    xval = float(splitline[0].strip())
                    yval = float(splitline[1].strip())
                    if xval == xvalPrev and yval == yvalPrev:   # avoid duplicate, dirty coordinates
                        logger.warning ("Airfoil '%s' has duplicate coordinates - skipped." % self._name)
                    else: 
                        x.append (xval)
                        y.append (yval) 
                    xvalPrev = xval 
                    yvalPrev = yval 
            else: 
                name = line.strip()
        
        if not name or not x or not y:
            raise ValueError ("Invalid .dat file")
        
        x, y = self._ensure_counter_clockwise (np.asarray (x), np.asarray (y))

        # test orientation 
        # x2, y2 = self._ensure_counter_clockwise (np.flip(np.copy (x)), np.flip(np.copy (y)))

        return name, x, y


    def _ensure_counter_clockwise (self, x : np.ndarray, y : np.ndarray):
        """ ensure x,y coordinates are counter clockwise ordered"""

        # sanity 

        if len(x) != len(y) or len(x) == 0:
            raise ValueError ("Invalid coordinates")

        # using shoelace formula to get signed area 
        #   A = 0.5 * sum (xi * yi+1 - xi+1 * yi )

        x_plus = np.append (x[1:], x[0])                            # shift index plus 1
        y_plus = np.append (y[1:], y[0])

        a = np.sum (x*y_plus) - np.sum(x_plus*y)

        if a < 0:                                                   # clockwise is negative 
            x, y = np.flip(x), np.flip(y)  
            logger.warning (f"{self} coordinates flipped to become counter clockwise")

        return x, y


    def save (self, onlyShapeFile=False):
        """
        Basic save of self to its pathFileName_abs
            for Hicks-Henne and Bezier 'onlyShapeFile' will write no .dat file 
        """
        if self.isLoaded: 
            self._write_dat_to_file ()
            self.set_isModified (False)

            logger.debug (f"{self} save to {self.fileName}")


    def saveAs (self, dir : str = None, destName : str = None, isWorkingDir = False):
        """
        save self to dir and destName and set new values to self
        if both dir and name are not set, it's just a save to current directory

        If workingDir is set, the saved airfoil will be relative to workingDir

        Returns: 
            newPathFileName from dir and destName 
        """     
        if destName: 
            self.set_name (destName)
            fileName = destName +  Airfoil.Extension
        else: 
            fileName = self.fileName  

        # create dir if not exist - build new airfoil filename
        if dir: 
            if not os.path.isdir (dir):
                os.mkdir(dir)
            if isWorkingDir:
                self.set_pathFileName (self.fileName)
                self.set_workingDir   (dir)
            else:
                self.set_pathFileName (os.path.join (dir, fileName), noCheck=True)

        self.save()
        self.set_isModified (False)
        return self.pathFileName



    def save_copyAs (self, dir = None, destName = None, te_gap=None ):
        """
        Write a copy of self to destPath and destName (the airfoil can be renamed).
        Self remains with its current values.
        Optionally a new te_gap may be defined for the exported airfoil  

        Args: 
            dir: -optional- new directory for the airfoil 
            destName: - optional- new name
            te_gap: -optional- new TE gap in x,y coordinates 

        Returns: 
            newPathFileName from dir and destName 
        """        

        # determine (new) airfoils name  if not provided
        if not destName:
            if self.name: 
                destName = self.name    
            else:
                if self.fileName:
                    destName = Path(self.fileName).stem        # cut '.dat'
                else: 
                    raise ValueError ("Destination name of airfoil couldn't be evaluated")

        # create dir if not exist - build airfoil filename
        if dir: 
            if not os.path.isdir (dir):
                os.mkdir(dir)
            newPathFileName = os.path.join (dir, destName) + Airfoil.Extension
        else: 
            newPathFileName = destName + Airfoil.Extension

        # create temp new airfoil 
        if not self.isLoaded: self.load()

        airfoil = self.asCopy (name=destName, pathFileName=newPathFileName)

        if te_gap is not None: 

            airfoil.geo.set_te_gap (te_gap)

            # build new airfoil name 
            mods = airfoil.geo.modifications_as_label
            airfoil.set_name (f"{airfoil._name_org}{mods}")
            airfoil.set_fileName_from_name ()

        # save it to file 
        airfoil.save ()

        return airfoil.pathFileName


    def asCopy (self, pathFileName = None, 
                name=None, nameExt=None,
                geometry=None) -> 'Airfoil':
        """
        returns a copy of self 

        Args:
            pathFileName: optional - string of existing (relative) airfoil path and name 
            name: optional         - name of airfoil - no checks performed 
            nameExt: -optional     - will be appended to self.name (if name is not provided)
            geometry: optional     - the geometry staretegy either GEO_BASIC, GEO_SPLNE...
        """
        if pathFileName is None and name is None: 
            pathFileName = self.pathFileName

        if os.path.isabs (pathFileName):
            workingDir = None
        else: 
            workingDir = self.workingDir 

        if name is None:
            name = self.name + nameExt if nameExt else self.name

        geometry = geometry if geometry else self._geometry_class

        airfoil =  Airfoil (x = np.copy (self.x), y = np.copy (self.y), 
                            name = name, pathFileName = pathFileName, 
                            workingDir = workingDir,
                            geometry = geometry )
        return airfoil 


    def asCopy_design (self, pathFileName = None) -> 'Airfoil':
        """
        returns a copy of self - same as asCopy with additional properties
            of a DESIGN airfoil

        Args:
            pathFileName: optional - string new fileName
        """
        pathFileName = pathFileName if pathFileName else self.pathFileName 
        name         = self.name 
        geometry     = self._geometry_class

        airfoil = self.asCopy (pathFileName=pathFileName, name=name, geometry=geometry)

        airfoil.set_usedAs (self.usedAs)
        airfoil.geo._modification_dict = copy (self.geo._modification_dict)
        airfoil._name_org = self._name_org
        
        return airfoil 


    def _write_dat_to_file (self):
        """ writes .dat file of to self.pathFileName"""

        # ensure extension .dat (in case of Bezier) 
        pathFileName_abs =  os.path.splitext(self.pathFileName_abs)[0] + Airfoil.Extension

        with open(pathFileName_abs, 'w+') as file:
            file.write("%s\n" % self.name)
            for i in range (len(self.x)):
                file.write("%.7f %.7f\n" %(self.x[i], self.y[i]))
            file.close()


    def normalize (self, 
                   just_basic=False, 
                   mod_string=None):
        """
        Shift, rotate, scale airfoil so LE is at 0,0 and TE is symmetric at 1,y
        Returns True/False if normalization was done 

        Args:
            just_basic: will only normalize coordinates - not based on spline 
            mod_string: alternate modification identifier for fileName and name 
        """
        normalize_done = self.geo.normalize(just_basic=just_basic)

        if normalize_done: 

            if mod_string:
                fileName_stem = os.path.splitext(self._fileName_org)[0]
                fileName_ext  = os.path.splitext(self._fileName_org)[1]
                self.set_fileName (fileName_stem + mod_string + fileName_ext)  
                self.set_name     (f"{self._name_org}")

        return normalize_done  


    def do_blend (self, 
                  airfoil1 : 'Airfoil', airfoil2 : 'Airfoil', 
                  blendBy : float,
                  geometry_class = None ):
        """ 
        Blends self out of two airfoils to the left and right
            depending on the blendBy factor
        
        Args: 
            airfoil1, airfoil2: Airfoils to the left and right 
            blendBy: 0.0 equals to the left, 1.0 equals to the right airfoil  
            geometry_class: optional - geo strategy for blend - either GEO_BASIC or GEO_SPLINE
        """
    
        blendBy = clip (blendBy, 0.0, 1.0)

        # other geo strategy? 
        if not geometry_class is None and geometry_class != self._geometry_class:
            geo : Geometry = geometry_class (self.x, self.y)
            geo.blend (airfoil1.geo, airfoil2.geo, blendBy)

            self._handle_geo_changed (geo=geo) 
        else:
            self.geo.blend (airfoil1.geo, airfoil2.geo, blendBy)

        self.set_isBlendAirfoil (True)


    @property
    def flap_setter (self) -> 'Flap_Setter':
        """ proxy controller to flap self using Worker"""

        if self._flap_setter is None and not self.isFlapped: 
            self._flap_setter = Flap_Setter (self)
        return self._flap_setter 


    def do_flap (self):
        """ flap self based on current 'flapper' data"""

        if self._flap_setter is None: return 

        # run worker - read flapped airfoil 
        self.flap_setter.set_flap ()

        if self.flap_setter.airfoil_flapped:

            # flapping was successful - update geometry - will update self
            x,y        = self.flap_setter.airfoil_flapped.x, self.flap_setter.airfoil_flapped.y
            flap_angle = self.flap_setter.flap_angle

            self.geo.set_flapped_data (x,y, flap_angle)


#------------------------------------------------------

class Airfoil_Bezier(Airfoil):
    """ 

    Airfoil based on Bezier curves for upper and lower side 

    """

    isBezierBased       = True

    Extension           = ".bez"


    def __init__(self, name = None, pathFileName=None, workingDir= None,
                 cp_upper = None,
                 cp_lower = None):
        """
        Main constructor for new Bezier Airfoil

        Args:
            pathFileName: optional - string of existinng airfoil path and name 
            name: optional - name of airfoil - no checks performed 
        """
        super().__init__( name = name, pathFileName=None, workingDir=workingDir)

        self._pathFileName   = pathFileName         # after super() as checks would hit 
        self._geometry_class = Geometry_Bezier      # geometry startegy 
        self._isLoaded       = False                # bezier definition loaded? 

        if cp_upper is not None: 
            self.geo.upper.set_controlPoints (cp_upper)
        if cp_lower is not None: 
            self.geo.lower.set_controlPoints (cp_lower)

        # pathFileName must exist if no coordinates were given 

        if (pathFileName is not None) and  (cp_upper is None or cp_lower is None): 
            pathFileName = os.path.normpath(pathFileName)       # make all slashes to double back
            if os.path.isabs (pathFileName):
                checkPath = pathFileName
            else:
                checkPath = self.pathFileName_abs
            if not os.path.isfile(checkPath):
                raise ValueError ("Airfoil file '%s' does not exist. Couldn't create Airfoil" % checkPath)
 

    @staticmethod
    def onAirfoil (anAirfoil : Airfoil):
        """
        Alternate constructor for new Airfoil based on another airfoil 

        The new Bezier airfoil will just have a rough estimation for the Bezier curves,
        which have to be optimized with 'match bezier'
        """

        # new name and filename
        name = anAirfoil.name + '_bezier'
 
        # get estimated controlpoints for upper and lower 
        cp_upper = Side_Airfoil_Bezier.estimated_controlPoints (anAirfoil.geo.upper, 5)
        cp_lower = Side_Airfoil_Bezier.estimated_controlPoints (anAirfoil.geo.lower, 5)

        airfoil_new =  Airfoil_Bezier (name=name, cp_upper=cp_upper, cp_lower=cp_lower)

        # new pathFileName
        fileName_stem = anAirfoil.fileName_stem
        # fileName_ext  = anAirfoil.fileName_ext
        pathFileName  = os.path.join (anAirfoil.pathName, fileName_stem + '_bezier' + Airfoil_Bezier.Extension)
        airfoil_new.set_pathFileName (pathFileName, noCheck=True)
        airfoil_new.set_workingDir   (anAirfoil.workingDir)

        airfoil_new.set_isLoaded (True)

        return airfoil_new 


    @property
    def pathFileName_shape (self) -> str: 
        """ abs pathfileName of the Bezier definition file """
        if self.pathFileName_abs:  
            return os.path.splitext(self.pathFileName_abs)[0] + Airfoil_Bezier.Extension
        else: 
            return None 

    @property
    def isLoaded (self): 
        # overloaded
        return self._isLoaded
    def set_isLoaded (self, aBool: bool):
        self._isLoaded = aBool

    @property
    def geo (self) -> Geometry_Bezier:
        """ the geometry strategy of self"""
        if self._geo is None: 
            self._geo = self._geometry_class (onChange = self._handle_geo_changed)
        return self._geo

    @override
    @property
    def flap_setter (self) -> 'Flap_Setter':
        """ proxy controller to flap self using Worker"""
        # do not flap Bezier
        return None 


    def set_xy (self, x, y):
        """ Bezier - do nothing """

        # overloaded - Bezier curve in Geometry is master of data 
        pass


    def set_newSide_for (self, curveType, px,py): 
        """creates either a new upper or lower side in self"""
        self.geo.set_newSide_for (curveType, px,py)
        self.set_isModified (True)

    @property
    def x (self):
        # overloaded  - take from bezier 
        return self.geo.x

    @property
    def y (self):
        # overloaded  - take from bezier 
        return self.geo.y

    # -----------------

    def reset (self): 
        """ make child curves like thickness or camber invalid """
        self.geo._reset_lines()


    def load (self):
        """
        Overloaded: Loads bezier definition instead of .dat from file" 
        """    

        if os.path.isfile (self.pathFileName_abs):

            self.load_bezier(fromPath=self.pathFileName_abs)
        
            # get modfication datetime of file 

            ts = os.path.getmtime(self.pathFileName_abs)                # file modification timestamp of a file
            self._file_datetime = datetime.fromtimestamp(ts)            # convert timestamp into DateTime object


    def load_bezier (self, fromPath=None):
        """
        Loads bezier definition from file. 
        pathFileName must be set before or fromPath must be defined.
        Load doesn't change self pathFileName
        """    

        with open(fromPath, 'r') as file:            

            file_lines = file.readlines()

        # format of bezier airfoil file 
        # <airfoil name> 
        # Top Start
        # 0.0000000000000000 0.0000000000000000
        # ...
        # 1.0000000000000000 0.0000000000000000
        # Top End
        # Bottom Start
        # ...
        # Bottom End

        new_name = 'Bezier_Airfoil'                             # default name 
        self._geo = None                                        # reset geometry 

        try: 
            px, py = [], []
            for i, line in enumerate(file_lines):
                if i == 0:
                    new_name = line.strip()
                else: 
                    line = line.lower()
                    if "start" in line:
                        if "top" in line: 
                            side = Line.Type.UPPER
                        else:
                            side = Line.Type.LOWER 
                        px, py = [], []
                    elif "end" in line:
                        if not px : raise ValueError("Start line missing")
                        if "top"    in line and side == Line.Type.LOWER: raise ValueError ("Missing 'Bottom End'")  
                        if "bottom" in line and side == Line.Type.UPPER: raise ValueError ("Missing 'Bottom Top'") 
                        self.set_newSide_for (side, px,py)
                    else:     
                        splitline = line.strip().split()
                        if len(splitline) == 1:                        # couldn't split line - try tab as separator
                            splitline = line.strip().split("\t")
                        if len(splitline) >= 2:                     
                            px.append (float(splitline[0].strip()))
                            py.append (float(splitline[1].strip()))
        except ValueError as e:
            logger.error ("While reading Bezier file '%s': %s " %(fromPath,e )) 
            return  
         
        self._name = new_name
        self._isLoaded = True 

        logger.debug (f"Bezier definition for {self.name} loaded")

        return   


    @override
    def save (self, onlyShapeFile=False):
        """
        Basic save of self to its pathFileName
            for Hicks-Henne and Bezier 'onlyShapeFile' will write no .dat file 
        """
        if self.isLoaded: 
            self._write_bez_to_file ()
            if not onlyShapeFile:
                self._write_dat_to_file ()

            self.set_isModified (False)

            logger.debug (f"{self} save to {self.fileName}")



    def _write_bez_to_file (self):
        """ write Bezier data to bez file """
        #  .bez-format for CAD etc and 

        with open(self.pathFileName_shape, 'w+') as file:

            # airfoil name 
            file.write("%s\n" % self.name)

            file.write("Top Start\n" )
            for p in self.geo.upper.controlPoints:
                file.write("%13.10f %13.10f\n" %(p[0], p[1]))
            file.write("Top End\n" )

            file.write("Bottom Start\n" )
            for p in self.geo.lower.controlPoints:
                file.write("%13.10f %13.10f\n" %(p[0], p[1]))
            file.write("Bottom End\n" )

            file.close()


    def asCopy (self, pathFileName = None, 
                name=None, nameExt=None,
                geometry=None) -> 'Airfoil':
        """
        returns a copy of self 

        Args:
            pathFileName: optional - string of existinng airfoil path and name 
            name: optional         - name of airfoil - no checks performed 
            nameExt: -optional     - will be appended to self.name (if name is not provided)
            geometry: - not supported - 
        """
        # overloaded as Bezier needs a special copy, no other geometry supported

        if pathFileName is None and name is None: 
            pathFileName = self.pathFileName

        if os.path.isabs (pathFileName):
            workingDir = None
        else: 
            workingDir = self.workingDir 

        if name is None:
            name = self.name + nameExt if nameExt else self.name

        if geometry is not None and geometry != Geometry_Bezier: 
            raise ValueError (f"Airfoil_Bezier does not support new geometry {geometry}")

        airfoil =  Airfoil_Bezier (name = name, pathFileName = pathFileName,
                                   workingDir=workingDir, 
                                   cp_upper = self.geo.upper.controlPoints,
                                   cp_lower = self.geo.lower.controlPoints)
        airfoil.set_isLoaded (True)

        return airfoil 


#------------------------------------------------------

class Airfoil_Hicks_Henne(Airfoil):
    """ 

    Airfoil based on a seed airfoil and hicks henne bump (hh) functions for upper and lower side 

    """

    isHicksHenneBased  = True

    Extension           = ".hicks"


    def __init__(self, name = None, pathFileName=None, workingDir= None):
        """
        Main constructor for new Airfoil

        Args:
            pathFileName: optional - string of existinng airfoil path and name \n
            name: optional - name of airfoil - no checks performed 
        """
        super().__init__( name = name, pathFileName=pathFileName, workingDir=workingDir)

        self._geometry_class  = Geometry_HicksHenne  # geometry strategy 
        self._isLoaded       = False                # hicks henne definition loaded? 

    @property
    def pathFileName_shape (self) -> str: 
        """ abs pathfileName of the hh definition file """
        if self.pathFileName_abs:  
            return os.path.splitext(self.pathFileName_abs)[0] + Airfoil_Hicks_Henne.Extension
        else: 
            return None 

    @property
    def isLoaded (self): 
        # overloaded
        return self._isLoaded
    def set_isLoaded (self, aBool: bool):
        self._isLoaded = aBool


    @property
    def geo (self) -> Geometry_HicksHenne:
        """ the geometry strategy of self"""
        if self._geo is None: 
            self._geo = self._geometry_class (onChange = self._handle_geo_changed)
        return self._geo

    @override
    @property
    def flap_setter (self) -> 'Flap_Setter':
        """ proxy controller to flap self using Worker"""
        # do not flap Hicks-Henne
        return None 


    def set_xy (self, x, y):
        """ hh - do nothing """
        # overloaded - hh geometry is master of data 
        pass

    @property
    def x (self):
        # overloaded  - take from geometry hh 
        return self.geo.x

    @property
    def y (self):
        # overloaded  - take from geometry hh  (seed airfoil and hicks henne are added)
        return self.geo.y

    # -----------------

    def reset (self): 
        """ make child curves like thickness or camber invalid """
        self.geo._reset_lines()


    def load (self):
        """
        Overloaded: Loads hicks henne definition instead of .dat from file" 
        """    

        if os.path.isfile (self.pathFileName_abs):

            self.load_hh(fromPath=self.pathFileName_abs)

            # get modfication datetime of file 

            ts = os.path.getmtime(self.pathFileName_abs)                       # file modification timestamp of a file
            self._file_datetime = datetime.fromtimestamp(ts)            # convert timestamp into DateTime object


    def load_hh (self, fromPath=None):
        """
        Loads hicks henne definition from file. 
        """    

        name, seed_name, seed_x, seed_y, top_hhs, bot_hhs = self._read_hh_file (fromPath)

        self.set_hh_data (name, seed_name, seed_x, seed_y, top_hhs, bot_hhs)



    def set_hh_data (self, name, seed_name, seed_x, seed_y, top_hhs, bot_hhs): 
        """ set all data needed for a Hicks Henne airfoil"""

        self._name = name                       # don't use set_ (isModified)

        if seed_name and len(seed_x) and len(seed_y): 

            seed_foil = Airfoil (x=seed_x, y=seed_y, name=seed_name)

            if seed_foil.isLoaded: 
                self._geo = Geometry_HicksHenne (seed_foil.x, seed_foil.y)
                self._geo.upper.set_hhs (top_hhs)
                self._geo.lower.set_hhs (bot_hhs)

                self._isLoaded = True 
                logger.debug (f"Hicks Henne definition for {self.name} loaded")
            else: 
                logger.error (f"Hicks Henne seed airfoil {seed_name} couldn't be loaded ")
        else: 
            raise ValueError (f"Hicks Henne seed airfoil data missing for {name}")



    def _read_hh_file (self, fromPath):
        """
        reads hicks henne definition from file. 
        """    

        from base.spline import HicksHenne

        with open(fromPath, 'r') as file:            

            file_lines = file.readlines()

        
        # format of bezier airfoil file 

        # <airfoil name> 
        # Top Start
        # 0.000strength000000000 0.0000location0000000  0.0000width0000000
        # ...
        # Top End
        # Bottom Start
        # ... 
        # Bottom End
        # Seedfoil Start 
        # 'seed airfoil name'
        #  1.000000 0.000000
        #  ...      ...

        name = ''                                # name of airfoil  
        seed_name = ''                           # name of seed airfoil 
        x,y = [], []                             # x,y of sedd
        top_hhs = []                             # array of hh functions 
        bot_hhs = []
        side = None

        try: 
            hhs = []
            for i, line in enumerate(file_lines):

                line_low = line.lower()

                if i == 0: 

                    name = line.strip()

                elif "seedfoil start" in line_low:

                    seed_name, x, y = self._loadLines (file_lines [i+1:])

                elif "start" in line_low:

                    if "top" in line_low: 
                        side = Line.Type.UPPER
                    else:
                        side = Line.Type.LOWER 
                    hhs = []

                elif "end" in line_low:

                    if not side : raise ValueError("Start line missing")
                    if "top"    in line_low and side == Line.Type.LOWER: raise ValueError ("Missing 'Bottom End'")  
                    if "bottom" in line_low and side == Line.Type.UPPER: raise ValueError ("Missing 'Bottom Top'") 

                    if side == Line.Type.LOWER:
                        bot_hhs = hhs
                    else: 
                        top_hhs = hhs 
                    side = None

                else:     

                    splitline = line.split()
                    if len(splitline) == 3:                     
                        strength = float(splitline[0].strip())
                        location = float(splitline[1].strip())
                        width    = float(splitline[2].strip())
                        hhs.append (HicksHenne (strength, location, width ))
        except ValueError as e:
            logger.error ("While reading Hicks Henne file '%s': %s " %(fromPath,e ))   
         
        return name, seed_name, x, y, top_hhs, bot_hhs   



#--------------------------------------------------------------------------


class Flap_Definition:
    """ 

    Defines the geometry of a flap 

    With set_flap a flapped version of the original airfoil is returned   

    """

    @staticmethod
    def have_same_hinge (flap_def1 : 'Flap_Definition', flap_def2 : 'Flap_Definition') -> bool:
        """
        Compare 2 flap definitions if they have the same hinge definition
        Return True if they are the same or both have no flap_def1
        """
        if flap_def1 and flap_def2:
            return  flap_def1.x_flap == flap_def2.x_flap and \
                    flap_def2.y_flap == flap_def2.y_flap and \
                    flap_def1.y_flap_spec == flap_def2.y_flap_spec
        elif flap_def1 is None and flap_def2 is None:
            return True
        else: 
            return False
        

    def __init__(self, dataDict : dict = None):
        """
        """

        self._x_flap        = fromDict (dataDict, "x_flap", 0.75)
        self._y_flap        = fromDict (dataDict, "y_flap", 0.0) 
        self._y_flap_spec   = fromDict (dataDict, "y_flap_spec", 'y/t')
        self._flap_angle    = fromDict (dataDict, "flap_angle", 0.0) 


    def _as_dict (self):
        """ returns a data dict with the parameters of self """

        d = {}
        toDict (d, "x_flap",        self.x_flap)                  
        toDict (d, "y_flap",        self.y_flap) 
        toDict (d, "y_flap_spec",   self.y_flap_spec) 
        toDict (d, "flap_angle",    self.flap_angle) 
        return d

    @property
    def name_suffix (self) -> str:
        """ 
        fileName suffix for being flapped like 
            '_f5.1' for defaults or 
            '_f-1.4_xf0.72_yf0.5_yspecYC' for non default values
        """

        return Worker.flapped_suffix (self.flap_angle, self.x_flap, self.y_flap, self.y_flap_spec)


    @property
    def x_flap (self) -> float: 
        return self._x_flap

    def set_x_flap (self, aVal : float):
        self._x_flap = clip (aVal, 0.02, 0.98)

    @property
    def y_flap (self) -> float: 
        return self._y_flap

    def set_y_flap (self, aVal : float):
        self._y_flap = clip (aVal, 0.0, 1.0)

    @property
    def y_flap_spec (self) -> str: 
        return self._y_flap_spec

    def set_y_flap_spec (self, aVal : str):
        self._y_flap_spec = aVal if aVal in ['y/c', 'y/t'] else self._y_flap_spec

    @property
    def flap_angle (self) -> float: 
        return self._flap_angle

    def set_flap_angle (self, aVal : float):
        self._flap_angle = clip (aVal, -20.0, 20.0)



class Flap_Setter (Flap_Definition):
    """ 

    Proxy to flap an airfoil using Worker 

    With set_flap a flapped version of the original airfoil is returned   

    """

    def __init__(self, airfoil_base : Airfoil):
        """
        constructor for new Flapper to handle flapping of an airfoil

        Args:
            airfoil_base:   an unflapped airfoil to flap 
        """

        super().__init__()
        
        if airfoil_base.isBezierBased or airfoil_base.isHicksHenneBased:
            raise ValueError ("Only .dat files can be flapped")
        
        if airfoil_base.isFlapped:
            raise ValueError ("A flapped airfoil cannot be flapped")
        
        self._worker_workingDir  = airfoil_base.pathName_abs        # working dir of Worker!
        self._base_copy          = airfoil_base.asCopy ()           # copy as parent will change!

        self._airfoil_flapped    = None                             # flapped version of airfoil_org 


    @property
    def airfoil_base (self) -> Airfoil:
        """ the initial, unflapped airfoil"""
        return self._base_copy

    @property
    def airfoil_flapped (self) -> Airfoil:
        """ airfoil org being flapped - None if not """
        return self._airfoil_flapped


    def set_flap (self, flap_angle=None, outname : str=None) -> Airfoil:
        """ 
        flap the base airfoil using worker - an optional flap angle can be submitted
        If successful, airfoil_flapped is available 
        """

        self._airfoil_flapped = None                        # reset flapped airfoil 

        if not Worker.ready:
            raise RuntimeError ("Worker is not ready to flap airfoil") 

        # don't do anything for flap angle = 0 

        if flap_angle is not None: 
            self.set_flap_angle (flap_angle)
        if self.flap_angle == 0.0: return

        # run Worker 

        worker = Worker(self._worker_workingDir)

        flapped_fileName = worker.set_flap (self.airfoil_base.fileName, 
                                        x_flap = self.x_flap, y_flap = self.y_flap, y_flap_spec = self.y_flap_spec,
                                        flap_angle = self.flap_angle,
                                        outname = outname )
        if flapped_fileName: 

            # load new airfoil 
            self._airfoil_flapped = Airfoil (pathFileName=flapped_fileName, workingDir=self._worker_workingDir)
            self._airfoil_flapped.load()

            # ... and delete immediatly its file to have a clean directory  
            try: 
                os.remove(self._airfoil_flapped.pathFileName_abs) 
            except OSError as exc: 
                logger.error (f"{self._airfoil_flapped.pathFileName_abs} couldn't be removed")




# ------------ test functions - to activate  -----------------------------------

if __name__ == "__main__":

    # # test flap set 

    # from airfoil_examples import Example

    # Worker().isReady (Path.cwd(), min_version='1.0.5')

    # airfoil = Example(geometry = GEO_SPLINE) 
    # # airfoil.geo.repanel (nPanels=300)
    # airfoil.save() 

    # flapper = Flap_Setter (airfoil)
    # for angle in np.arange (-10,10, 1.0):
    #     flapper.set_flap (angle) 
    # for y in np.arange (0,1, 0.1):
    #     flapper.set_y_flap (y)
    #     flapper.set_flap (15, outname=f"{airfoil.fileName}_y{y:.1f}") 
    # for x in np.arange (0,1, 0.1):
    #     flapper.set_x_flap (x)
    #     flapper.set_flap (15, outname=f"{airfoil.fileName}_x{x:.1f}") 
    
    pass  



#!/usr/bin/env pythonupper
# -*- coding: utf-8 -*-

"""  

The "Artists" to plot a airfoil object on a pg.PlotItem 

"""
import html 

from typing                     import Callable

from base.artist                import *
from base.common_utils          import *

from model.airfoil              import Line
from model.polar_set            import * 
from model.xo2_input            import OpPoint_Definition, OpPoint_Definitions, OPT_TARGET, OPT_MAX, OPT_MIN, GeoTarget_Definitions
from model.xo2_results          import OpPoint_Result, Xo2_Results, Optimization_History_Entry, GeoTarget_Result

from airfoil_artists            import _color_airfoil

from PyQt6.QtGui                import QColor, QBrush, QPen, QPainterPath, QTransform
from PyQt6.QtCore               import pyqtSignal


import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


# -------- helper functions ------------------------

def _size_opPoint (weighting : float, normal_size = 13) -> float:
    """ returns plot size of an opPoint (def) depending on weighting"""

    # weighting can be negative (meaning fixed weighting - no dynamic)
    return abs(weighting) ** 0.7 * normal_size 



# -------- Movable Points  ------------------------


class Movable_Xo2_OpPoint_Def (Movable_Point):
    """ 
    Movable opPoint definition
    """

    name = "OpPoint Def"

    COLOR = QColor ("darkturquoise")

    # -- special symbols for min/max optTypes 
    #
    #               .--x
    #               |
    #               y
    #

    @staticmethod
    def _symbol_opt_up (width=1.0, height1=0.3, height2=0.5):
        path = QPainterPath()

        path.addRect(QtCore.QRectF(-width / 2, -height1 / 2, width, height1))
        # vertcial line
        path.moveTo(0.0, -height1 / 2)
        path.lineTo(0.0, -height2 * 2)
        #triangle
        path.moveTo(-width / 2, -height2 * 2)
        path.lineTo( width / 2, -height2 * 2)
        path.lineTo(       0.0, -height2 * 2 - height2)
        path.lineTo(-width / 2, -height2 * 2)

        return path

    tr : QTransform = QTransform()
    tr.rotate(-90)                                  # because y is downward

    SYMBOL_OPT_UP    = _symbol_opt_up ()                       
    SYMBOL_OPT_LEFT  = tr.map (SYMBOL_OPT_UP)                         
    SYMBOL_OPT_DOWN  = tr.map (SYMBOL_OPT_LEFT)                        
    SYMBOL_OPT_RIGHT = tr.map (SYMBOL_OPT_DOWN)                        


    def __init__ (self, pi : pg.PlotItem, 
                    opPoint_def : OpPoint_Definition,
                    xyVars : Tuple[var, var], 
                    movable = True, 
                    on_selected = None,
                    on_dblClick = None,
                    on_delete   = None,
                    **kwargs):

        self._pi = pi
        self._opPoint_def = opPoint_def
        self._xyVars = xyVars

        self._callback_selected = on_selected
        self._callback_dblClick = on_dblClick
        self._callback_delete   = on_delete
        self._highlight_item    = None                       

        brush = QColor ("black")
        brush.setAlphaF (0.3) 

        size = _size_opPoint (opPoint_def.weighting)

        super().__init__(self._point_xy(), movable=movable, label_anchor = (0, 0.5),
                            color=QColor(self.COLOR), brush=brush,
                            symbol=self._symbol(), size=size, show_label_static = True, **kwargs)

        self.sigShiftClick.connect  (self._delete_opPoint_def)


    def _symbol (self) -> QPainterPath | str:
        """ returns the symbol for self depending on optType"""

        xVar, yVar = self._xyVars

        optVar, optType = self._opPoint_def.optVar_type_for_xyVars (self._xyVars)

        if optType == OPT_TARGET:
            s = 's'
        elif optType == OPT_MAX:
            if optVar == yVar:
                s = self.SYMBOL_OPT_UP
            else:
                s = self.SYMBOL_OPT_RIGHT
        elif optType == OPT_MIN:
            if optVar == yVar:
                s = self.SYMBOL_OPT_DOWN
            else:
                s = self.SYMBOL_OPT_LEFT
        else: 
            s = 'star'
        
        return s


    def _delete_opPoint_def (self, _):
        """ shift-click - delete me opPoint_dev"""

        if callable(self._callback_delete):
            # callback / emit signal delayed so we leave the scope of Graphics 
            QTimer() .singleShot(10, lambda: self._callback_delete (self._opPoint_def))   


    def _point_xy (self) -> tuple:  
        return self._opPoint_def.xyValues_for_xyVars (self._xyVars)     


    @override
    def mouseClickEvent(self, ev):
        """ handle callback to parent when clicked with left button to select self """

        if not self.moving and ev.button() == QtCore.Qt.MouseButton.LeftButton  \
                           and not (ev.modifiers() & QtCore.Qt.KeyboardModifier.ShiftModifier): # shift is handled in super
            if callable(self._callback_selected):
                ev.accept()

                self._opPoint_def.set_as_current()

                # callback / emit signal delayed so we leave the scope of Graphics 
                QTimer() .singleShot(10, lambda: self._callback_selected ())   
        else:
            super().mouseClickEvent (ev)


    @override
    def mouseDoubleClickEvent(self, ev):
        """ handle callback to parent when double clicked """
        if callable(self._callback_dblClick):
            ev.accept()

            self._opPoint_def.set_as_current()

            # callback / emit signal delayed so we leave the scope of Graphics 
            QTimer() .singleShot(10, lambda: self._callback_dblClick ())   
        else:
            super().mouseDoubleClickEvent (ev)


    @override
    def _moving (self, _):
        """ slot -point is moved"""
        self._opPoint_def.set_xyValues_for_xyVars (self._xyVars, (self.x,self.y))

        self.setPos (self._point_xy())                      # ... if we run against limits 

        # also refresh pos of highlight item 
        if isinstance (self._highlight_item, pg.TargetItem):
            self._highlight_item.setPos (self.pos())


    @override
    def label_static (self, *_) -> str:
        """ the static label - can be overloaded """
        return f"{self._opPoint_def.iPoint}" 

    @override
    def label_moving (self, *_):
        """ label for hovering or moving"""
        return f"{self._opPoint_def.labelLong_for (self._xyVars, fix=False)}"

    @override
    def _finished (self, _):
        """ slot - point moving is finished"""

        self._opPoint_def.set_as_current()

        # callback / emit signal delayed so we leave the scope of Graphics 
        QTimer() .singleShot(10, lambda: self._callback_changed ())   


    def set_highlight_item (self, aItem: pg.TargetItem):
        """ a point item used to highlight current"""

        self._highlight_item = aItem
        if self._highlight_item:
            self._highlight_item.setPos (self.pos())




# -------- Artists ------------------------


class Xo2_OpPoint_Defs_Artist (Artist):
    """ Plot / Modify Xoptfoil2 operating point defin itions """

    sig_opPoint_def_changed     = pyqtSignal ()             # opPoint_def changed changed 
    sig_opPoint_def_selected    = pyqtSignal ()
    sig_opPoint_def_dblClick    = pyqtSignal ()

    def __init__ (self, *args, 
                  isRunning_fn = None, 
                  xyVars = (var.CD, var.CL), 
                  **kwargs):
        super().__init__ (*args, **kwargs)

        self._xyVars = xyVars                               # definition of x,y axis
        self._isRunning_fn = isRunning_fn


    @property
    def xyVars(self): return self._xyVars
    def set_xyVars (self, xyVars: Tuple[var, var]): 
        """ set new x, y variables for polar """
        self._xyVars = xyVars 
        self.refresh()


    @property
    def opPoint_defs (self) -> OpPoint_Definitions: 
        return self.data_list

    @property
    def opPoint_def_are_editable (self) -> bool:
        """ True if optimier is ready - definitions can be changed"""
        return not self._isRunning_fn() if callable (self._isRunning_fn) else True


    def _plot (self): 

        legend_name = "Op Point Definition" 

        for opPoint_def in self.opPoint_defs:

            x, y = opPoint_def.xyValues_for_xyVars (self.xyVars)

            # does it fit into this x,y polar view

            if x is not None and y is not None: 

                pt = Movable_Xo2_OpPoint_Def  (self._pi, opPoint_def, self.xyVars, 
                                                movable=self.opPoint_def_are_editable and self.show_mouse_helper,
                                                on_changed =self.sig_opPoint_def_changed.emit,
                                                on_delete  =self._on_delete,
                                                on_dblClick=self.sig_opPoint_def_dblClick.emit,
                                                on_selected=self.sig_opPoint_def_selected.emit)
                self._add (pt, name = legend_name) 

                legend_name = None                                  # legend only once 

                # highlight current opPoint def for edit with a big circle 

                if opPoint_def == self.opPoint_defs.current_opPoint_def and self.opPoint_def_are_editable:

                    brush = QColor (Movable_Xo2_OpPoint_Def.COLOR)
                    brush.setAlphaF (0.3)
                    highlight_item = self._plot_point (0.02,0.2, color="black", size=60, brush=brush)

                    pt.set_highlight_item (highlight_item)


        # make scene clickable to add wing section - delayed as during init scene is not yet available

        QTimer().singleShot (10, self._connect_scene_mouseClick)

        self.show_help_message ()


    def show_help_message (self):
        """ show mouse helper message"""
        msg = "OpPoint: click to select,  double-click to edit,  shift-click to remove,  ctrl-click to add"
        self.set_help_message (msg)


    def _connect_scene_mouseClick (self): 
        """ connect mouse click in scene to slot"""           

        scene : pg.GraphicsScene = self._pi.scene()
        if scene:  scene.sigMouseClicked.connect (self._scene_clicked)


    @override
    def _remove_plots(self):
        """ overloaded to disconnect older slot when doing refresh"""

        super()._remove_plots()

        scene : pg.GraphicsScene = self._pi.scene()

        try:                            # slot could be not connected up to now
            scene.sigMouseClicked.disconnect (self._scene_clicked)
        except:
            pass


    def _scene_clicked (self, ev : MouseClickEvent):
        """ 
        slot - mouse click in scene of self - handle add wing section with crtl-click 
        """ 

        # handle only ctrl-click
        if not (ev.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier): return  
       
        # was the scene click in my viewbox?
        if isinstance(ev.currentItem, pg.ViewBox):
            viewbox : pg.ViewBox = ev.currentItem
        else:
            viewbox : pg.ViewBox = ev.currentItem.getViewBox()

        if viewbox == self._pi.vb:                     # is this my view box (or of another plot item)? 
            
            # get scene coordinates of click pos and map to view box 
            ev.accept()
            pos : pg.Point = viewbox.mapSceneToView(ev.scenePos())

            # create new opPoint_def in the list of opPoint_defs 
            self.opPoint_defs.create_in_xyVars (self.xyVars, pos.x(), pos.y())

            self.sig_opPoint_def_changed.emit()


    def _on_delete (self, opPoint_def : OpPoint_Definition):
        """ callback of Movable Point to delete opPoint_def"""

        self.opPoint_defs.delete (opPoint_def)

        self.sig_opPoint_def_changed.emit ()


class Xo2_OpPoint_Artist (Artist):
    """ Plot Xoptfoil2 operating point results """

    def __init__ (self, *args, 
                  opPoint_defs_fn : Callable = None,
                  opPoint_results_fn : Callable = None,
                  prev_opPoint_results_fn : Callable = None,
                  xyVars = (var.CD, var.CL), 
                  **kwargs):

        self._xyVars = xyVars                                       # definition of x,y axis
        self._opPoint_defs_fn = opPoint_defs_fn                     # method to get opPoint definitions
        self._opPoint_results_fn = opPoint_results_fn               # method to get opPoint results
        self._prev_opPoint_results_fn = prev_opPoint_results_fn     # method to get previous opPoint results 

        super().__init__ (*args, **kwargs)

    @property
    def xyVars(self): return self._xyVars
    def set_xyVars (self, xyVars: Tuple[var, var]): 
        """ set new x, y variables for polar """
        self._xyVars = xyVars 
        self.refresh()


    def _plot (self): 
        """ plot all opPoints"""

        opPoints :      list[OpPoint_Result] = self._opPoint_results_fn() 
        if not opPoints: return 

        opPoint_defs :  list[OpPoint_Definition] = self._opPoint_defs_fn() 
        prev_opPoints : list[OpPoint_Result] = self._prev_opPoint_results_fn() [:] 

        legend_name = f"Op Point Result - Design {opPoints[0].idesign}"

        textFill = QColor ("black")
        textFill.setAlphaF (0.5)
       
        for iop, opPoint in enumerate (opPoints):

            prev_opPoint = prev_opPoints[iop] if prev_opPoints else None
            x = opPoint.get_value (self.xyVars[0])
            y = opPoint.get_value (self.xyVars[1])

            # sanity - opPoint def could be deleted in the meantime 
            opPoint_def = opPoint_defs[iop] if len(opPoint_defs) == len (opPoints) else None

            color  = self._opPoint_color  (opPoint, opPoint_def)
            symbol = self._opPoint_symbol (opPoint, prev_opPoint)
            label  = self._opPoint_label  (opPoint, opPoint_def)
            size   = _size_opPoint (opPoint.weighting)

            textColor = QColor (color).darker(130)

            self._plot_point ((x,y), color=color, size=size, symbol=symbol, zValue=20, name=legend_name,
                              text=label, anchor=(0, 0.5), textOffset = (8,0), textColor=textColor, textFill=textFill)

            legend_name = None 

        # message if dynamic weighting of opPoints occured 

        if prev_opPoints:
            weightings      = [op.weighting for op in opPoints]
            prev_weightings = [op.weighting for op in prev_opPoints]
            if sum(weightings) != sum (prev_weightings):
                self._plot_text ('New weightings applied', color= "dimgray", fontSize=self.SIZE_HEADER, itemPos=(0.5, 1))

        
    def _opPoint_color (self, opPoint : OpPoint_Result, opPoint_def : OpPoint_Definition | None) -> QColor:
        """ color of opPoint depending on % deviation """

        # sanity 
        if opPoint_def is None: 
            return COLOR_OK

        deviation = opPoint.deviation

        optType = opPoint_def.optType
        optVar  = opPoint_def.optVar
        allow_improved = opPoint_def._myList.allow_improved_target

        if optType == OPT_TARGET:                            # targets - take deviation to target
            # Xoptfoil2 - colors for deviation 
            #
            #   if (op_spec%allow_improved_target .and. opt_type == 'target-drag' .and. dev < 0d0) then 
            #     how_good = Q_GOOD
            #   else if (op_spec%allow_improved_target .and. opt_type /= 'target-drag' .and. dev > 0d0) then 
            #     how_good = Q_GOOD
            #   else
            #     how_good = r_quality (abs(dev), 0.1d0, 2d0, 10d0)      ! in percent
            #   end if
            if allow_improved   and optVar == var.CD and deviation < 0.0:
                color = COLOR_GOOD
            elif allow_improved and optVar != var.CD and deviation > 0.0:
                color = COLOR_GOOD
            else:
                deviation = abs(deviation)
                if deviation < 0.1:
                    color = COLOR_GOOD
                elif deviation >= 10:
                    color = COLOR_ERROR
                elif deviation >= 2.0:
                    color = COLOR_WARNING
                elif deviation >= 0.1:
                    color = COLOR_OK
                else:
                    color = COLOR_OK
        else:                                               # min/max - deviation is improvement 
            # Xoptfoil2 - colors for improvement 
            #
            # if (improv <= 0d0) then 
            #     how_good = Q_BAD
            # elseif (improv < 5d0) then 
            #     how_good = Q_OK
            # elseif (improv >= 5d0) then 
            #     how_good = Q_GOOD
            # else
            #     how_good = Q_BAD
            # end if 
            if optType == OPT_MAX:                          # eg. OPT_MAX of 'glide' - more is better 
                improv = deviation
            else:                                           # eg OPT_MIN of 'cd' - less is better
                improv = deviation

            if improv <= -0.1:
                color = COLOR_WARNING
            elif improv < 2.0:
                color = COLOR_OK
            elif improv >= 2.0:
                color = COLOR_GOOD
            else:
                color = COLOR_ERROR
        return color 


    def _opPoint_symbol (self, opPoint : OpPoint_Result, prev_opPoint : OpPoint_Result | None ):
        """ a triangle in the direction of value change """

        if prev_opPoint is None: return 'o'

        x = opPoint.get_value (self.xyVars[0])
        y = opPoint.get_value (self.xyVars[1])
        prev_x = prev_opPoint.get_value (self.xyVars[0])
        prev_y = prev_opPoint.get_value (self.xyVars[1])

        if prev_x > x:
            symbol = 't3'               # triangle to left
        elif prev_x < x:
            symbol = 't2'               # triangle to right
        elif prev_y > y:
            symbol = 't'                # triangle to down
        elif prev_y < y:
            symbol = 't1'               # triangle to up
        else: 
            symbol = 'o'
        return symbol

        
    def _opPoint_label (self, opPoint : OpPoint_Result, opPoint_def : OpPoint_Definition | None) -> str:
        """ label of opPoint depending on % deviation """

        label = None 

        # sanity 
        if opPoint_def is None: 
            return label

        distance  = opPoint.distance

        optType = opPoint_def.optType
        optVar  = opPoint_def.optVar
        allow_improved = opPoint_def._myList.allow_improved_target

        if optType == OPT_TARGET:                            # targets - take deviation to target

            if allow_improved   and optVar == var.CD and distance < 0.0:
                distance = 0.0
            elif allow_improved and optVar != var.CD and distance > 0.0:
                distance = 0.0
            else:
                distance = abs(distance)

            if optVar == var.CD and round_down (distance,5):
                label = f"delta {optVar}: {distance:.5f}"
            elif optVar != var.CD and round_up (distance,2):
                label = f"delta {optVar}: {distance:.2f}"        
            else: 
                label = f"hit"        

        else:                                               # min/max - deviation is improvement 

            if optType == OPT_MAX:                          # eg. OPT_MAX of 'glide' - more is better 
                improv = distance
            else:                                           # eg OPT_MIN of 'cd' - less is better
                improv = distance

            if improv: 
                if optVar == var.CD:
                    label = f"d{var.CD} {improv:.5f}"
                else:
                    label = f"d{optVar} {improv:.2f}"
        return label 
        



class Xo2_Design_Radius_Artist (Artist):
    """ Plot Xoptfoil2 design radius during optimization """

    @property
    def results (self) -> Xo2_Results: 
        return self.data_object
    
    @property
    def steps (self) -> list[Optimization_History_Entry]: 
        """ optimization step entries up to now """
        return self.results.steps  

    def _plot (self): 

        radius_list = [entry.design_radius for entry in self.steps]

        if radius_list:
            p = pg.PlotDataItem  (radius_list)
            self._add (p)

            # plot last as point 
            x = len(radius_list) - 1
            y = radius_list[-1]
            back_color = QColor ("black")
            back_color.setAlphaF (0.5)

            self._plot_point (x,y, size=5, color=COLOR_OK, text = f"{y:.3f}", textFill=back_color, anchor= (1.1, 0.9))



class Xo2_Improvement_Artist (Artist):
    """ Plot Xoptfoil2 improvement during optimization """

    @property
    def results (self) -> Xo2_Results: 
        return self.data_object
    
    @property
    def steps (self) -> list[Optimization_History_Entry]: 
        """ optimization step entries up to now """
        return self.results.steps  

    def _plot (self): 

        improvement_list = [entry.improvement for entry in self.steps]

        if improvement_list:

            p = pg.PlotDataItem  (improvement_list)
            self._add (p)

            # plot last as point 
            x = len(improvement_list) - 1
            y = improvement_list[-1]
            back_color = QColor ("black")
            back_color.setAlphaF (0.5)
            
            self._plot_point (x,y, size=5, color=COLOR_GOOD, text = f"{y:.2f}%", textFill=back_color, 
                              textOffset=(-5,0), anchor= (0.8, -0.1))




class Xo2_Transition_Artist (Artist):
    """ Plot Xoptfoil2 point of transition on design airfoil """

    # ----------  Symbol transition - create an arrow up and rotate to right 
    #
    #               .--x
    #               |
    #               y

    @staticmethod
    def _symbol_transition_right ():
        path = QPainterPath()
        # coords = [(-0.125, 0.125), (0, 0), (0.125, 0.125),
        #             (0.05, 0.125), (0.05, 0.5), (-0.05, 0.5), (-0.05, 0.125)]
        coords = [
                (0,0),
                (0,0.125),
                (-0.04, 0.125),
                (-0.04, -0.125),
                (0, -0.125),
                (0,0),
                (0.05, 0.075),
                (0.15, -0.075),
                (0.25, 0.075),
                (0.35, -0.075),
                (0.4, 0)]
        path.moveTo(*coords[0])
        for x,y in coords[1:]:
            path.lineTo(x, y)
        # path.closeSubpath()
        return path
    # tr : QTransform = QTransform()
    # tr.rotate(90)                                   # because y is downward
    # tr.translate (0,-0.5)                          # translate is after rotation ...

    # SYMBOL_TRANSITION_RIGHT  = tr.map (_symbol_transition_up())                         
    SYMBOL_TRANSITION_RIGHT  = _symbol_transition_right()                         

    # ----------------

    def __init__ (self, *args, 
                  opPoints_result_fn : Callable = None,
                  **kwargs):

        self._opPoints_result_fn = opPoints_result_fn                  # method to get opPoints of designs

        super().__init__ (*args, **kwargs)


    def _plot (self): 
        """ plot all opPoints"""

        airfoil : Airfoil = self.data_object
        opPoints_result : list[OpPoint_Result ]= self._opPoints_result_fn() 

        if not opPoints_result or not airfoil: return 

        legend_name = "Point of Transition xtr"

        color  = _color_airfoil ([], airfoil).darker (120) 
        symbol = self.SYMBOL_TRANSITION_RIGHT
        size   = 40
        brush  = QColor ("black")
        fill   = QColor ("black")
        fill.setAlphaF (0.5) 

        for side in [airfoil.geo.lower, airfoil.geo.upper]:     

            for iop, opPoint in enumerate (opPoints_result):

                x = opPoint.xtrt if side.isUpper else opPoint.xtrb
                y = side.yFn (x) 
                y = y + 0.008 if side.isUpper else y - 0.008

                text   = f"{iop+1}"
                anchor = (0.5, 1.2) if side.isUpper else (0.5, -0.2)

                self._plot_point (x,y, color=color, symbol=symbol, size=size, brush=brush, 
                                  text=text, textColor=color, anchor=anchor, textFill=fill, name=legend_name)

                legend_name = None                                      # only once for legend 


    @override
    def _add_legend_item (self, plot_item, name : str = None):
        """ add legend item having 'name'"""

        # non standard legend size of transition symbol 
        if self._pi.legend is not None and self.show_legend and name:

            if isinstance (plot_item,  pg.TargetItem):

                # create a dummy PlotItem as TargetItem won't appear in legend 
                size_legend = 30    
                pen    = plot_item.pen
                symbol = plot_item._path 
                brush  = plot_item.brush
                p = pg.ScatterPlotItem ([], [], pen= pen, brush=brush, symbol=symbol, size=size_legend, pxMode=True)
                self._add (p, name=name) 

            else: 

                super()._add_legend_item (plot_item, name=name)
     

class Xo2_GeoTarget_Artist (Artist):
    """ Plot Xoptfoil2 geometry target results """

    def __init__ (self, *args, 
                  geoTarget_defs_fn : Callable = None,
                  geoTarget_results_fn : Callable = None,
                  **kwargs):

        self._geoTarget_defs_fn    = geoTarget_defs_fn                     # method to get geo target definitions
        self._geoTarget_results_fn = geoTarget_results_fn                  # method to get geo target results

        super().__init__ (*args, **kwargs)


    def _plot (self): 
        """ plot all geo target results"""

        geoTargets :      list[GeoTarget_Result] = self._geoTarget_results_fn() 
        if not geoTargets: return 

        geoTarget_defs :  GeoTarget_Definitions = self._geoTarget_defs_fn() 
       
        for igeo, geoTarget_def in enumerate (geoTarget_defs):

            for geoTarget in geoTargets:                                    # get both definition and result
                if geoTarget_def.optVar == geoTarget.optVar:

                    target_text = f"Target {geoTarget_def.optVar}: {geoTarget_def.optValue:.2%}"

                    if round (geoTarget.deviation,1) != 0.0:
                        result_text = f"delta: {geoTarget.distance:.2%}"
                    else: 
                        result_text = f"hit"        

                    deviation = abs (round (geoTarget.deviation,1))
                    if deviation < 0.1:
                        result_color = COLOR_GOOD
                    elif deviation >= 10:
                        result_color = COLOR_ERROR
                    elif deviation >= 2.0:
                        result_color = COLOR_WARNING
                    elif deviation >= 0.1:
                        result_color = COLOR_OK
                    else:
                        result_color = COLOR_OK

                    self._plot_text (target_text, parentPos=(0.5,0.1), itemPos=(1.0, 0.5), offset=(0, igeo*20))
                    self._plot_text (result_text, parentPos=(0.5,0.1), itemPos=(0.0, 0.5), offset=(5, igeo*20), color=result_color)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Math utility functions

"""

import numpy as np
import math
from bisect import bisect_left

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.WARNING)


#------------ time to run -----------------------------------


    # from timeit import default_timer as timer
    # start = timer()
    # ...
    # print("Time ", timer() - start)  



#------------ Point -----------------------------------

from typing import overload

class JPoint:
    """    
    Simple point class with basic functionality

    'J' because there are so many Point classes ...
    """

    def __init__ (self, a, b = None, 
                  x_limits : tuple|None = None, 
                  y_limits : tuple|None = None ,
                  name : str|None = None):
        """ 
        Create new point. The arguments a,b can be 
            - x, y cordinates 
            - (x,y,) tuple of coordinates
            - Point (x,y) 
        """

        self._x = None 
        self._x_limits = x_limits 

        self._y = None
        self._y_limits = y_limits 

        self._name = name 
        self._fixed = False

        self.set_xy (a, b)

        logger.debug (f"{self} new")


    def __repr__(self) -> str:
        # overwritten to get a nice print string 
        text = f"({self.x},{self.y})"
        return f"<{type(self).__name__} {text}>"

    @property
    def x (self) -> float:
        return self._x

    @property
    def y (self) -> float:
        return self._y

    @property
    def xy (self) -> tuple[float]:
        return (self.x, self.y)

    @property
    def x_limits (self) -> tuple:
        return self._x_limits

    @property
    def y_limits (self) -> tuple:
        return self._y_limits

    @property
    def name (self):
        return self._name 
    def set_name (self, aStr: str):
        self._name = aStr

    def label_percent (self) -> str:
        """ returns a short label 7.4@30.6 for y,x values as percent""" 

        lab = f"{self._y*100:.1f}@{self._x*100:.1f}"
        return lab 

    def set_x_limits (self, lim : tuple): 
        """ set new x coordinate limits"""
        self._x_limits = lim
 
    def set_y_limits (self, lim : tuple): 
        """ set new y coordinate limits"""
        self._y_limits = lim
 
    @property
    def fixed (self) -> bool:
        """ True if x and y are fixed """
        return self._fixed
    
    def set_fixed (self, fixed : bool):
        """ set if x and y are fixed """
        self._fixed = fixed is True 


    def set_x (self, x : float): 
        """ set new x coordinate"""
        self._x = self._set_val (self._x, x, self._x_limits, self._fixed)
       
    def set_y (self, y : float): 
        """ set new y coordinate"""
        self._y = self._set_val (self._y, y, self._y_limits, self._fixed)


    def set_xy (self, a, b=None):
        """ 
        Set new point coordinates. The arguments a, b can be 
            - x, y cordinates 
            - (x,y,) tuple of coordinates
            - Point (x,y) 
        """ 
        if isinstance (a, int): a = float(a)
        if isinstance (b, int): b = float(b)

        if isinstance (a, float) and isinstance (b, float):
            x_new = a 
            y_new = b 
        elif isinstance (a, tuple):
            x_new = float (a[0])
            y_new = float (a[1])
        elif isinstance (a, JPoint):
            x_new = a.x, 
            y_new = a.y 
        else:
            raise ValueError (f"Point - cannot init with {a}, {b}")
        
        self._x = self._set_val (self._x, x_new, self._x_limits, self._fixed)
        self._y = self._set_val (self._y, y_new, self._y_limits, self._fixed)


    def isNew (self, x_new, y_new, decimals=7) -> tuple[bool]:
        """ returns True/False for x_new and y_new if they are different from current x,y,"""

        if x_new:
            x_isNew = round(x_new,decimals) != round (self.x, decimals)
        else: 
            x_isNew = False 

        if y_new:
            y_isNew = round(y_new,decimals) != round (self.y, decimals)
        else: 
            y_isNew = False 

        return x_isNew, y_isNew


    def as_transformed (self, transform_fn) -> 'JPoint':
        """ 
        Transform self with a transformation function transform_fn and
        returns a new JPoint
            The function must accept x,y and has to return xt, yt
            as the new coordinate values.
            If transform_fn is None, return self.
        """

        # transform coordinate 

        if transform_fn is None: 
            return self
        else: 
            if not callable (transform_fn):
                raise ValueError (f"transformation function is not callable{transform_fn}")
            xt, yt = transform_fn (self.x, self.y)

        # transform limits

        if self.x_limits is not None:
            x_min, x_max =  self.x_limits
            xt_min,_ = transform_fn (x_min, self.y)
            xt_max,_ = transform_fn (x_max, self.y)
            if xt_min <= xt_max:                                    # transform may flip limits 
                xt_limits = (xt_min, xt_max)
            else: 
                xt_limits = (xt_max, xt_min)
        else: 
            xt_limits = None

        if self.y_limits is not None:
            y_min, y_max =  self.y_limits
            _, yt_min = transform_fn (self.x, y_min)
            _, yt_max = transform_fn (self.x, y_max)
            if yt_min <= yt_max:                                    # transform may flip limits 
                yt_limits = (yt_min, yt_max)
            else: 
                yt_limits = (yt_max, yt_min)
        else: 
            yt_limits = None

        jPoint =  JPoint (xt, yt, x_limits=xt_limits, y_limits=yt_limits, name=self.name)
        jPoint.set_fixed (self.fixed)

        return jPoint


    def _set_val (self, val, new_val, limits, is_fixed) -> float:
        """  set new_val of val - return new_val"""

        if is_fixed:
            new_val = val 
        else: 
            new_val = float(new_val)

            min_val = limits [0] if limits is not None else new_val 
            max_val = limits [1] if limits is not None else new_val

            new_val = max (new_val, min_val)
            new_val = min (new_val, max_val)

            new_val = round (new_val,10)            # avoid float issues 

        return new_val 

    @staticmethod
    def transform (jpoints : list['JPoint'], transform_fn) -> list['JPoint']:
        """ transforms a list of JPoints and returns them as a new list"""

        jpoints_transformed = []
        for jpoint in jpoints:
            jpoints_transformed.append (jpoint.as_transformed (transform_fn))
        return jpoints_transformed
    

#------------ linear interpolation -----------------------------------


def interpolate(x1:float, x2:float, y1:float, y2:float, x:float) -> float:

    if (x1 -x) == 0.0: return y1
    
    try:
        y = ((y2-y1)/(x2-x1)) * (x-x1) + y1
    except:
        raise ValueError (f"Division by zero in interpolation , x1: {x1}, x2: {x2}")
    return y



#------------ rounding -----------------------------------


def round_up (number : float, decimals : int =0) -> float:
    """ round number up with to deciamls"""

    multiplier = 10 ** decimals
    multiplied = round(number * multiplier, 1)        # avoid Python float artefacts 
    return math.ceil(multiplied) / multiplier



def round_down (number : float, decimals : int =0) -> float:
    """ round number down with to deciamls"""

    multiplier = 10 ** decimals
    multiplied = round(number * multiplier, 1)        # avoid Python float artefacts 
    return math.floor(multiplied) / multiplier



#------------ panel angles -----------------------------------


def panel_angles (x,y):
    """returns an array of panel angles of polyline x,y - between 160 - 180
    angle[0] and [-1] default to 180° 
    """

    # Xfoil - CANG 

    # C---- go over each point, calculating corner angle
    #       IF(IPRINT.EQ.2) WRITE(*,1050)
    #       DO 30 I=2, N-1
    #         DX1 = X(I) - X(I-1)
    #         DY1 = Y(I) - Y(I-1)
    #         DX2 = X(I) - X(I+1)
    #         DY2 = Y(I) - Y(I+1)
    # C
    # C------ allow for doubled points
    #         IF(DX1.EQ.0.0 .AND. DY1.EQ.0.0) THEN
    #          DX1 = X(I) - X(I-2)
    #          DY1 = Y(I) - Y(I-2)
    #         ENDIF
    #         IF(DX2.EQ.0.0 .AND. DY2.EQ.0.0) THEN
    #          DX2 = X(I) - X(I+2)
    #          DY2 = Y(I) - Y(I+2)
    #         ENDIF
    # C
    #         CROSSP = (DX2*DY1 - DY2*DX1)
    #      &         / SQRT((DX1**2 + DY1**2) * (DX2**2 + DY2**2))
    #         ANGL = ASIN(CROSSP)*(180.0/3.1415926)
    #         IF(IPRINT.EQ.2) WRITE(*,1100) I, X(I), Y(I), ANGL
    #         IF(ABS(ANGL) .GT. ABS(AMAX)) THEN
    #          AMAX = ANGL
    #          IMAX = I
    #         ENDIF
    #    30 CONTINUE

    angles = np.zeros (len(x))
    for i in range(1, len(x)-1):
        dx1 = x[i] - x[i-1] 
        dy1 = y[i] - y[i-1] 
        dx2 = x[i] - x[i+1] 
        dy2 = y[i] - y[i+1] 
        if dx1 != 0.0 and dx2 != 0.0:               # check for pathologic airfoil (blunt le) 
            crossp = (dx2 * dy1 - dy2 * dx1) / math.sqrt ((dx1**2 + dy1**2) * (dx2**2 + dy2**2))
            angles[i] = math.asin(crossp)
        else: 
            angles[i] = 0.0
    angles = 180.0 - angles * (180/np.pi)
    return angles 



#------------ derivative -----------------------------------


def derivative1 (x, y):
        """
        evaluate first derivative of polyline (x,y)
        using Backward, center, forward adapted difference approximation
        """

        #     npt = size(x)        
        #     do i = 1, npt
        #       if (i == 1) then                                                 ! forward
        #         derivative1(i) = (y(i+1) - y(i))/ (x(i+1) - x(i)) 
        #       else if (i ==npt) then                                           ! backward
        #         derivative1(i) = (y(i) - y(i-1))/ (x(i) - x(i-1))
        #       else                                                             ! center
        #         h       = x(i+1) - x(i)
        #         h_minus = x(i) - x(i-1)
        #         hr      = h / h_minus 
        #         derivative1(i) = (y(i+1) - hr*hr*y(i-1) -(1-hr*hr)*y(i))/ (h * (1.d0 +hr)) 
        #       end if 
        #     end do

        npt = len(x)
        der1 = np.zeros (npt)
        for i in range(len(x)):

            if i == 0: 
                h = x[i+1] - x[i]
                if h != 0.0:
                    der1[i] = (y[i+1] - y[i]) / h
                else: 
                    der1[i] = None

            elif i == npt-1:
                h_minus = x[i] - x[i-1]
                if h_minus != 0.0:
                    der1[i] = (y[i] - y[i-1]) / h_minus
                else: 
                    der1[i] = None

            else: 
                h = x[i+1] - x[i]
                h_minus = x[i] - x[i-1]
                if h != 0.0 and h_minus != 0.0:
                    hr = h / h_minus
                    der1[i] = (y[i+1] - hr**2 * y[i-1] -(1 - hr**2) * y[i]) / (h * (1+hr))
                else: 
                    der1[i] = None

        return der1
                     



#------------ find closest   -----------------------------------


def find_closest_index (myList, myNumber) -> int:
    """
    Assumes myList is sorted. Returns index of closest value to myNumber.

    If two numbers are equally close, return index of smallest number.
    """
    pos = bisect_left(myList, myNumber)
    if pos == 0:
        return 0
    if pos == len(myList):
        return -1
    before = myList[pos - 1]
    after = myList[pos]
    if after - myNumber < myNumber - before:
        return pos
    else:
        return pos - 1


#------------ Bisection - find index  -----------------------------------

def bisection(array,value):
    '''Given an ``array`` , and given a ``value`` , returns an index j such that ``value`` is between array[j]
    and array[j+1]. ``array`` must be monotonic increasing. j=-1 or j=len(array) is returned
    to indicate that ``value`` is out of range below and above respectively.'''

    # from https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
    
    n = len(array)
    if (value < array[0]):
        return -1
    elif (value > array[n-1]):
        return n
    jl = 0                          # Initialize lower
    ju = n-1                        # and upper limits.
    while (ju-jl > 1):              # If we are not yet done,
        jm=(ju+jl) >> 1             # compute a midpoint with a bitshift
        if (value >= array[jm]):
            jl=jm                   # and replace either the lower limit
        else:
            ju=jm                   # or the upper limit, as appropriate.
        # Repeat until the test condition is satisfied.
    if (value == array[0]):         # edge cases at bottom
        return 0
    elif (value == array[n-1]):     # and top
        return n-1
    else:
        return jl


#------------ Bisection - find Root  -----------------------------------


def bisection_fn (f,a,b,N, tolerance=None):
    '''Approximate solution of f(x)=0 on interval [a,b] by bisection method.

    Parameters
    ----------
    f : function
        The function for which we are trying to approximate a solution f(x)=0.
    a,b : numbers
        The interval in which to search for a solution. The function returns
        None if f(a)*f(b) >= 0 since a solution is not guaranteed.
    N : (positive) integer
        The number of iterations to implement.

    Returns
    -------
    x_N : number
        The midpoint of the Nth interval computed by the bisection method. The
        initial interval [a_0,b_0] is given by [a,b]. If f(m_n) == 0 for some
        midpoint m_n = (a_n + b_n)/2, then the function returns this solution.
        If all signs of values f(a_n), f(b_n) and f(m_n) are the same at any
        iteration, the bisection method fails and return None.
    '''
    # from https://patrickwalls.github.io/mathematicalpython/root-finding/root-finding/

    if   abs(f(a)) < 10e-10 : return a, 0
    elif abs(f(b)) < 10e-10 : return b, 0
    elif f(a)*f(b) > 0:
        print("Bisection method fails.", a, f(a), b, f(b))
        return None, 0
    a_n = a
    b_n = b
    for n in range(1,N+1):
        m_n = (a_n + b_n)/2
        if not tolerance is None and abs(a_n - b_n) < tolerance:
            return m_n, n 
        f_m_n = f(m_n)
        if f(a_n)*f_m_n < 0:
            a_n = a_n
            b_n = m_n
        elif f(b_n)*f_m_n < 0:
            a_n = m_n
            b_n = b_n
        elif f_m_n == 0:
            print("Found exact solution.")
            return m_n, n
        else:
            print("Bisection method fails.")
            return None

    return (a_n + b_n)/2, n


def secant_fn (f,a,b,N):
    '''Approximate solution of f(x)=0 on interval [a,b] by the secant method.

    Parameters
    ----------
    f : function
        The function for which we are trying to approximate a solution f(x)=0.
    a,b : numbers
        The interval in which to search for a solution. The function returns
        None if f(a)*f(b) >= 0 since a solution is not guaranteed.
    N : (positive) integer
        The number of iterations to implement.

    Returns
    -------
    m_N : number
        The x intercept of the secant line on the the Nth interval
            m_n = a_n - f(a_n)*(b_n - a_n)/(f(b_n) - f(a_n))
        The initial interval [a_0,b_0] is given by [a,b]. If f(m_n) == 0
        for some intercept m_n then the function returns this solution.
        If all signs of values f(a_n), f(b_n) and f(m_n) are the same at any
        iterations, the secant method fails and return None.

    Examples
    --------
    >>> f = lambda x: x**2 - x - 1
    >>> secant(f,1,2,5)
    1.6180257510729614
    '''

    # from https://patrickwalls.github.io/mathematicalpython/root-finding/secant/

    if f(a)*f(b) >= 0:
        print("Secant method fails.")
        return None
    a_n = a
    b_n = b
    for n in range(1,N+1):
        m_n = a_n - f(a_n)*(b_n - a_n)/(f(b_n) - f(a_n))
        f_m_n = f(m_n)
        if f(a_n)*f_m_n < 0:
            a_n = a_n
            b_n = m_n
        elif f(b_n)*f_m_n < 0:
            a_n = m_n
            b_n = b_n
        elif f_m_n == 0:
            # print("Found exact solution.")
            return m_n, n
        else:
            print("Secant method fails.")
            return None
    return a_n - f(a_n)*(b_n - a_n)/(f(b_n) - f(a_n)), n



#------------ Newton iteration - find Root  -----------------------------------

def newton(f,Df,x0,epsilon = 10e-8 , max_iter= 50, bounds=None):
    '''Approximate solution of f(x)=0 by Newton's method.

        Implement Newton's method: compute the linear approximation
        of f(x) at xn and find x intercept by the formula
            x = xn - f(xn)/Df(xn)
        Continue until abs(f(xn)) < epsilon and return xn.
        If Df(xn) == 0, raise Error .

            Parameters
    ----------
    f : function
        Function for which we are searching for a solution f(x)=0.
    Df : function
        Derivative of f(x).
    x0 : number
        Initial guess for a solution f(x)=0.
    epsilon : number
        Stopping criteria is abs(f(x)) < epsilon.
    max_iter : integer
        Maximum number of iterations of Newton's method.
    bounds : optional - tuple of lower and upper bound of x

    Returns
    -------
    xn : number
    niter : iterations needed 
    '''
    # from https://patrickwalls.github.io/mathematicalpython/root-finding/root-finding/

    xn = x0
    for n in range(0,max_iter):

        if bounds is not None: 
            if   xn > bounds[1]: xn = bounds[1]
            elif xn < bounds[0]: xn = bounds[0]

        fxn = f(xn)
        if abs(fxn) < epsilon:
            break

        Dfxn = Df(xn)
        if Dfxn == 0:
            if xn == 0.0:                       # special case at LE 
                break
            else: 
                raise ValueError ("Newton iteration: Zero derivative. No solution found.")
        xn = xn - fxn/Dfxn

    if bounds is not None: 
        if   xn > bounds[1]: xn = bounds[1]
        elif xn < bounds[0]: xn = bounds[0]

    return xn, n



# ---------------------------------------------------------------------------
# (c) https://github.com/fchollet/nelder-mead 
# 
# Pure Python/Numpy implementation of the Nelder-Mead optimization algorithm
# to determine the minimum of a fuction - replaces fmin and brentq from scipy


def nelder_mead_1D (f, x_start,
                    
                step=0.1, no_improve_thr=10e-10,                # for scalar product - org: no_improve_thr=10e-8,
                no_improv_break=12, max_iter=50,
                bounds = None,                                  # extension 
                alpha=1., gamma=2., rho=-0.5, sigma=0.5):
    '''
        1D Nelder-Mead optimization algorithm to determine the minimum of a fuction

        1D version for speed optimization in simple cases 

        Parameters
        ----------
        f : function to optimize, must return a scalar score
        x_start : float - initial position
        step    : float - look-around radius in initial step
        no_improv_thr: float - an improvement lower than no_improv_thr
        no_improv_break : int -break after no_improv_break iterations 
        bounds : tuple(float) - (min, max) pair for boundary of x. None - no boundary. 
             
        Returns
        -------
        xbest : x of minimum  
        score : best score  
        niter : int - iterations needed 
        """

    '''

    def penalty (x, bounds):
        if bounds is None: return 0.0
        
        if x < bounds[0] or x > bounds[1]:
            return 9999.9
        else:
            return 0.0

    def fn_penalty (f, x, bounds):
        # return function value - if x outside bounds return penality value 
        if bounds is None: return f(x)
        
        if penalty (x, bounds): 
            fn = penalty (x, bounds)  
        else: 
            fn = f(x)
        return fn


    # init
    prev_best = f(x_start)
    no_improv = 0
    res = [(x_start, prev_best)]

    # mod 
    if penalty(x_start + step, bounds) > 0:
        x = x_start - step 
    else: 
        x = x_start + step

    score = fn_penalty (f, x, bounds) 
    res.append((x, score))

    # simplex iter
    iters = 0
    while 1:
        # order
        res.sort(key=lambda score: score[1])
        best = res[0][1]

        # break after max_iter
        if max_iter and iters >= max_iter:
            return res[0][0], res[0][1], iters
        iters += 1

        # break after no_improv_break iterations with no improvement
        # print ('...best so far:', best)

        if abs(best - prev_best) < no_improve_thr:
            no_improv += 1
        else:
            no_improv = 0
            prev_best = best

        # if iters % 20 == 0: 
        #     print("NelderMead  - %3d(%d) score: %.5f  no_improv: %2d thresh: %.5f" %(iters, max_iter, best, no_improv, no_improve_thr))

        if no_improv >= no_improv_break:
            return res[0][0], res[0][1], iters

        # centroid
        x0 = 0.0
        for tup in res[:-1]:
            x0 += tup[0] / (len(res)-1)

        # reflection
        xr = x0 + alpha*(x0 - res[-1][0])
        rscore = fn_penalty (f, xr, bounds)

        if res[0][1] <= rscore < res[-2][1]:
            del res[-1]
            res.append((xr, rscore))
            continue

        # expansion
        if rscore < res[0][1]:
            xe = x0 + gamma*(x0 - res[-1][0])
            escore = fn_penalty (f, xe, bounds)

            if escore < rscore:
                del res[-1]
                res.append((xe, escore))
                continue
            else:
                del res[-1]
                res.append((xr, rscore))
                continue

        # contraction
        xc = x0 + rho*(x0 - res[-1][0])
        cscore = fn_penalty (f, xc, bounds)
        if cscore < res[-1][1]:
            del res[-1]
            res.append((xc, cscore))
            continue

        # reduction
        x1 = res[0][0]
        nres = []
        for tup in res:
            redx = x1 + sigma*(tup[0] - x1)
            score = fn_penalty (f, redx, bounds)
            nres.append((redx, score))
        res = nres




def nelder_mead (f, 
                x_start,
                step : float = 0.1, 
                no_improve_thr : float = 10e-10,                          # for scalar product
                no_improv_break : int = 12, 
                no_improv_break_beginning : int|None = None, 
                max_iter : int =50,
                bounds = None,                                   
                alpha=1., gamma=2., rho=-0.5, sigma=0.5,
                stop_callback : bool =False):
    '''
        Nelder-Mead optimization algorithm to determine the minimum of a fuction

        Allows multi dimensional optimization of scalar function f. 
        For 1D use 'nelder_mead_1D' which should be faster  

        Arguments:

        f       : function to optimize, must return a scalar score
        x_start : initial position
        step    : look-around radius in initial step
        no_improv_thr: an improvement lower than no_improv_thr
        no_improv_break : break after no_improv_break iterations 
        no_improv_break_beginning : break after no_improv_break iterations at the beginning (default=no_improv_break) 
        bounds  : list of tuple(float) - (min, max) pair for boundary of x. None - no boundary. 
        max_iter: always break after this number of iterations.
        alpha, gamma, rho, sigma (floats): parameters of the algorithm (see Wikipedia page for reference)
        stop_callback : optional method for stop condition
             
        Returns:
         
        (xbest,score) : np array tuple - x of minimum  , best score 
        niter : int - iterations needed 
        """

    '''
    def penalty (x, bounds):
        if bounds is None: return 0.0
        
        if x < bounds[0] or x > bounds[1]:
            return 9999.9
        else:
            return 0.0

    def fn_penalty (f, x, bounds):
        # return function value - if x outside bounds return penality value 
        if bounds is None: return f(x)
        
        for i in range (len(x)): 
            if penalty (x[i], bounds[i]): 
                return penalty (x[i], bounds[i])
            
        return f(x)

    # sanity

    if stop_callback and not callable (stop_callback):
        stop_callback = False

    # init

    dim = len(x_start)
    if bounds is None: 
        bounds = [None] * dim                         # dummy bounds for all x

    prev_best = f(x_start)
    no_improv = 0

    if no_improv_break_beginning is None: 
        no_improv_break_beginning = no_improv_break


    res = [[np.copy(x_start), prev_best]]

    for i in range(dim):
        x = np.copy(x_start)
        if penalty (x[i] + step, bounds[i]):
            x[i] = x[i] - step
        else:
            x[i] = x[i] + step
        score = fn_penalty (f, x, bounds) 
        res.append([x, score])

    # simplex iter
    iters = 0
    # xr, xe, xc = 0, 0, 0 

    while 1:
        # order
        res.sort(key=lambda x: x[1])
        best = res[0][1]

        # stop request?
        if stop_callback(): 
            return res[0], iters
        
        # break after max_iter
        if max_iter and iters >= max_iter:
            return res[0], iters
        iters += 1

         # break after no_improv_break iterations with no improvement
 
        if abs(best - prev_best) < no_improve_thr:
            no_improv += 1
        else:
            no_improv = 0
            prev_best = best

        # allow a different (higher) no_improv_break at the beginning (results are still highly volatile) 
        if iters < max_iter / 5:
            if no_improv >= no_improv_break_beginning:
                # print(f"  nelder_mead iters: {iters} best: {best} prev_best {prev_best} no_improv: {no_improv}") 
                return res[0], iters
        else: 
            if no_improv >= no_improv_break:
                # print(f"  nelder_mead iters: {iters} best: {best} prev_best {prev_best} no_improv: {no_improv}") 
                return res[0], iters


        # centroid
        x0 = [0.0] * dim 
        for tup in res[:-1]:
            for i, c in enumerate(tup[0]):
                x0 [i] += c / (len(res)-1)
        # print ("Centroid ", iters, x0)
        # for i, tup in enumerate(res):
        #     print ("res ", i, tup[0])
        # reflection
        xr = x0 + alpha*(x0 - res[-1][0])
        rscore = fn_penalty (f, xr, bounds) 

        if res[0][1] <= rscore < res[-2][1]:
            del res[-1]
            res.append([xr, rscore])
            continue

        # expansion
        if rscore < res[0][1]:
            xe = x0 + gamma*(x0 - res[-1][0])
            escore = fn_penalty (f, xe, bounds) 
            if escore < rscore:
                del res[-1]
                res.append([xe, escore])
                continue
            else:
                del res[-1]
                res.append([xr, rscore])
                continue

        # contraction
        xc = x0 + rho*(x0 - res[-1][0])
        cscore = fn_penalty (f, xc, bounds) 
        if cscore < res[-1][1]:
            del res[-1]
            res.append([xc, cscore])
            continue

        # reduction
        x1 = res[0][0]
        nres = []
        for tup in res:
            redx = x1 + sigma*(tup[0] - x1)
            score = fn_penalty (f, redx, bounds) 
            nres.append([redx, score])
        res = nres


#--- wrapper functions for nelder_mead minimum------

def nelder_mead_wrap  (fn, xStart,
                no_improve_thr=10e-10,               
                no_improv_break=10, # 5, 
                max_iter=50,
                bounds = None): 
    
    if not bounds is None:
        if xStart < bounds[0] or xStart > bounds[1]:
            raise ValueError ("nelder-mead: Start value %.6f outside bounds" % xStart)
        
        step = (bounds[1] - bounds[0]) / 15.01 #**10.01
    else: 
        step = 0.051

    
    xmin, score, niters =  nelder_mead_1D(fn, xStart, step= step,
                                          no_improve_thr=no_improve_thr, 
                                          no_improv_break=no_improv_break, max_iter=max_iter,  
                                          bounds=bounds)    
    # print ("nelder_mead : ", xmin, score, niters)
    
    # workaround of bug in melder-mead
    # seems to be a float issue as it happens at rounded decimals like 0.61 or 0.8
    # retry with a new xStart-Value solves ... most of the time.

    if niters < max_iter and score > no_improve_thr:
        if not bounds is None: 
            # move start towards bounds0
            xStart_new =  bounds[0] + (xStart - bounds[0]) * 0.9
            # old - xStart_new = (xStart + bounds[0])  / 1.8 # airfoil 1.8 # planform 1.99
            step_new   = (bounds[1] - bounds[0]) / 20.01 # 5.01 
            if xStart_new < bounds[0] or xStart > bounds[1]:
                raise ValueError ("nelder-mead: Start value %.6f outside bounds" % xStart_new)
        else: 
            xStart_new = xStart *  1.012
            step_new   = 0.011
        # try longer 
        no_improv_break = no_improv_break * 3
        # print ("nelder_mead bug:   ", xmin, score, niters, xStart, step, bounds)
        xmin, score, niters =  nelder_mead_1D(fn, xStart_new, step=step_new, 
                                            no_improve_thr=no_improve_thr, 
                                            no_improv_break=no_improv_break, max_iter=max_iter,  
                                            bounds=bounds)    
        # print ("nelder_mead retry: ", xmin, score, niters)
        if score > no_improve_thr:
            raise ValueError ("nelder-mead: Minimum not found for xStart = %.4f" % xStart_new)

    return xmin


def findMin (fn, xStart, bounds=None, no_improve_thr=10e-10) -> float: 
    xmin =  nelder_mead_wrap (fn, xStart, bounds=bounds, no_improve_thr=no_improve_thr)    
    return xmin 

def findMax (fn, xStart, bounds=None) -> float: 
    xmax =  nelder_mead_wrap(lambda x: - (fn(x)), xStart, bounds=bounds)    
    return xmax 

def findRoot (fn, xStart, bounds=None, no_improve_thr=10e-12) -> float: 
    xRoot =  nelder_mead_wrap(lambda x: abs(fn(x)), xStart, no_improve_thr=no_improve_thr,  bounds=bounds)    
    return xRoot 





if __name__ == "__main__":

    a = JPoint (1,3)    
    a = JPoint ((1,3))
    a = JPoint (JPoint(1,3))
    b = JPoint (1) 



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""  

The "Artists" to plot a airfoil object on a pg.PlotItem 

"""
import html 

from typing                     import Callable

from base.math_util             import derivative1
from base.artist                import *
from base.common_utils          import *
from base.spline                import Bezier, HicksHenne

from model.airfoil              import Airfoil, Airfoil_Bezier, usedAs, Geometry, Flap_Setter
from model.airfoil_geometry     import Line, Side_Airfoil_Bezier, Side_Airfoil_HicksHenne
from model.polar_set            import * 

from PyQt6.QtGui                import QColor, QBrush, QPen
from PyQt6.QtCore               import pyqtSignal, QObject


import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


# -------- helper functions ------------------------

def _color_airfoil (airfoils : list[Airfoil], airfoil: Airfoil) -> QColor:
    """ returns QColor for airfoil depending on its type """

    alpha = 1.0
    airfoil_type = airfoil.usedAs

    if airfoil_type == usedAs.DESIGN:
        color = 'deeppink'
    elif airfoil_type == usedAs.NORMAL:
        color = 'springgreen'  
    elif airfoil_type == usedAs.FINAL:
        # color = QColor('turquoise').lighter (120)
        color = 'springgreen'  
    elif airfoil_type == usedAs.SEED:
        color = 'dodgerblue'
    elif airfoil_type == usedAs.SEED_DESIGN:
        color = 'cornflowerblue'
    elif airfoil_type == usedAs.REF:                         
        color = _color_airfoil_ref (airfoils, airfoil)
        # alpha = 0.9
    elif airfoil_type == usedAs.TARGET:
        color = 'cornflowerblue'
    elif airfoil_type == usedAs.SECOND:
        color = 'cornflowerblue'
    else:
        color = 'gray'
    qcolor =  QColor (color) 

    if alpha != 1.0:
        qcolor.setAlphaF (alpha) 
    return qcolor


def _color_airfoil_ref (airfoils : list[Airfoil], airfoil: Airfoil) -> QColor:
    """ returns QColor for reference airfoil """

    # get the how many reference 
    iRef, nRef = 0, 0 
    for a in airfoils:
        if a == airfoil: iRef = nRef 
        if a.usedAs == usedAs.REF: nRef += 1

    # get color depending on iRef 
    color = color_in_series ('lightskyblue', iRef, nRef, delta_hue=0.4)
    # color.setAlphaF (0.9)

    return color 


def _label_airfoil (airfoils : list[Airfoil], airfoil: Airfoil) -> str:
    """ nice label including usedAs for airfoil"""

    if airfoil.usedAs == usedAs.REF:
        iRef, nRef = 0, 0                                       # get the how many reference 
        for a in airfoils:
            if a == airfoil: iRef = nRef 
            if a.usedAs == usedAs.REF: nRef += 1
        use = f"Ref {iRef+1}: "
    elif airfoil.usedAs == usedAs.DESIGN or airfoil.usedAs == usedAs.NORMAL or airfoil.usedAs is None: # no prefix 
        use = ""
    else: 
        use = f"{airfoil.usedAs}: " 

    label = f"{use}{airfoil.name_to_show}"
    return label 


def _linestyle_of (aType : Line.Type) -> QColor:
    """ returns PenStyle for line depending on its type """

    if aType == Line.Type.CAMBER:
        style = style=Qt.PenStyle.DashDotLine
    elif aType == Line.Type.THICKNESS:
        style = style=Qt.PenStyle.DashLine
    elif aType == Line.Type.UPPER:
        style = style=Qt.PenStyle.DotLine
    elif aType == Line.Type.LOWER:
        style = style=Qt.PenStyle.DotLine
    else:
        style = style=Qt.PenStyle.SolidLine

    return style


class Movable_Highpoint (Movable_Point):
    """ 
    Represents the highpoint of an airfoil Line object,
    which can be moved to change the highpoint 
    """

    def __init__ (self, 
                  geo : Geometry, 
                  line : Line, 
                  line_plot_item : pg.PlotDataItem, *args, 
                  movable : bool = False,
                  color : QColor|str, 
                  **kwargs):

        # symmetrical and camber? 
        if line.type == Line.Type.CAMBER and geo.isSymmetrical: 
            movable = False 
        # Bezier is changed via control points ? 
        elif geo.isBezier:
            movable = False 

        self.name = 'max '+ line.name
        self._geo = geo
        self._line = line
        self._line_type = line.type
        self._line_plot_item = line_plot_item

        super().__init__(line.highpoint.xy, 
                         color = color, 
                         movable = movable, 
                         show_label_static = movable,
                         **kwargs)


    def label_static (self, *_):

        if self._line.type == Line.Type.CAMBER and self._geo.isSymmetrical: 
            return  "No camber - symmetrical" 
        else:  
            return super().label_static()

    def label_moving (self, *_):

        if self._line.type == Line.Type.CAMBER and self._geo.isSymmetrical: 
            return  "No camber - symmetrical" 
        else:  
            return f"{self.y:.2%} @ {self.x:.2%}"


    def _moving (self, _):
        """ slot - point is moved by mouse """
        # overlaoded to update airfoil geo 
        # update highpoint coordinates
        self._geo.set_highpoint_of (self._line, (self.x, self.y), finished=False)

        # update self xy if we run against limits 
        self.setPos(self._line.highpoint.xy)

        # update line plot item 
        self._line_plot_item.setData (self._line.x, self._line.y)


    def _finished (self, _):
        """ slot - point is move finished """
        # overlaoded to update airfoil geo 
        self._geo.finished_change_of (self._line)

        # update parent plot item 
        self._changed()




class Movable_TE_Point (Movable_Point):
    """ 
    Represents the upper TE point. 
    When moved the upper plot item will be updated
    When finished final TE gap will be set 
    """

    name = "TE gap"

    def __init__ (self, 
                  geo : Geometry, 
                  upper_plot_item : pg.PlotDataItem, 
                  movable = False, 
                  **kwargs):

        # Bezier is changed via control points ? 
        if geo.isBezier:
            movable = False 

        self._geo = geo
        self._upper_plot_item = upper_plot_item

        xy = self._te_point_xy()

        super().__init__(xy, movable=movable, show_label_static = movable,**kwargs)

    
    def _te_point_xy (self): 
        return self._geo.upper.x[-1], self._geo.upper.y[-1]
    

    def _moving (self, _):
        """ slot -point is moved"""

        # update highpoint coordinates
        self._geo.set_te_gap (self.y * 2 , moving=True)

        # update self xy if we run against limits 
        self.setPos(self._te_point_xy())

        # update line plot item 
        self._upper_plot_item.setData (self._geo.upper.x, self._geo.upper.y)


    def _finished (self, _):
        """ slot - point moving is finished"""
        # final highpoint coordinates
        self._geo.set_te_gap (self.y * 2, moving=False)
        self._changed()


    def label_moving (self, *_):

        return f"{self.name}  {self.y*2:.2%} "

    def _label_opts (self, moving=False, hover=False) -> dict:
        """ returns the label options as dict """

        # overloaded to align right 
        if moving or hover:
            labelOpts = {'color': QColor(Artist.COLOR_NORMAL),
                        'anchor': (1,1),
                        'offset': (10, 10)}
        else: 
            labelOpts = {'color': QColor(Artist.COLOR_LEGEND),
                        'anchor': (1,1),
                        'offset': (10, 10)}
        return labelOpts




class Movable_LE_Point (Movable_Point):
    """ 
    Represents the LE radius . 
    When moved the LE radius will be updated
    When finished final LE radius will be set 
    """
    name =  "LE radius"

    def __init__ (self, 
                  geo : Geometry, 
                  circle : pg.ScatterPlotItem, 
                  movable = False, 
                  **kwargs):

        # Bezier is changed via control points ? 
        if geo.isBezier:
            movable = False 

        self._geo = geo
        self._circle_item = circle
        xy = 2 * self._geo.le_radius , 0

        super().__init__(xy, movable=movable, show_label_static = movable, **kwargs)
  

    def _moving (self, _):
        """ slot - when point is moved"""

        # update radius 
        new_radius = self.x / 2
        new_radius = min(0.05,  new_radius)
        new_radius = max(0.002, new_radius)

        # update self xy if we run against limits 
        self.setPos(2 * new_radius,0)

        # update line plot item 
        self._circle_item.setData ([new_radius], [0])
        self._circle_item.setSize (2 * new_radius)


    def _finished (self, _):
        """ slot - point moving is finished"""

        # final highpoint coordinates
        new_radius = self.x / 2
        self._geo.set_le_radius (new_radius)
        self._changed()


    def label_moving (self, *_):

        return f"{self.name}  {self.x/2:.2%} "





class Movable_Side_Bezier (Movable_Bezier):
    """
    pg.PlotCurveItem/UIGraphicsItem which represents 
    an airfoil Side_Bezier. 
    The Bezier curve which can be changed by the controllpoints
    
    Points are implemented with Moveable_Points
    A Moveable_Point can also be fixed ( movable=False).
    See pg.TargetItem for all arguments 

    Callback 'on_changed' will return the (new) list of 'points'

    """
    def __init__ (self, 
                  airfoil : Airfoil_Bezier,
                  side : Side_Airfoil_Bezier,
                  **kwargs):

        self._airfoil = airfoil
        self._side = side 
        points = side.controlPoints_as_points

        if side.isUpper:
            label_anchor = (0,1) 
        else: 
            label_anchor = (0,0)

        super().__init__(points, label_anchor=label_anchor, **kwargs)

    @property
    def bezier (self) -> Bezier:
        """ the Bezier  self is working with """
        # here - take Bezier of the 'side' 
        return self._side.bezier

    @property
    def u (self) -> list:
        """ the Bezier parameter  """
        # here - take Bezier of the 'side' 
        return self._side._u


    def refresh (self):
        """ refresh control points from side control points """

        # update all my movable points at once 
        movable_point : Movable_Bezier_Point
        for i, point_xy in enumerate(self._side.controlPoints): 
            movable_point = self._movable_points[i]
            movable_point.setPos_silent (point_xy)              # silent - no change signal 

        self.setData(*self.points_xy())                         # update self (polyline) 

        # if self._bezier_item is not None:   	                # update bezier item to be dispalyed
        #     x,y = self.bezier.eval (self.u) 
        #     self._bezier_item.setData (x, y)
        #     self._bezier_item.show()


    def scene_clicked (self, ev : MouseClickEvent):
        """ 
        slot - mouse click in scene of self 
            - handle add Bezier point with crtl-click either on upper or lower side
        """ 

        # handle on ctrl-click
        if not (ev.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier): return  
       
        # get scene coordinates of click pos and map to view box 
        vb : pg.ViewBox = self.getViewBox()
        pos : pg.Point = vb.mapSceneToView(ev.scenePos())
        pos_x = pos.x()
        pos_y = pos.y()

        # typically there are 2 instances of self - upper and lower Bezier 
        if pos_y < 0.0 and self._side.isLower:
            self._add_point (pos_x, pos_y)
        elif pos_y >= 0.0 and self._side.isUpper:
            self._add_point (pos_x, pos_y)


    def _add_point (self, pos_x, pos_y):
       """ add controlpoint to Bezier curve""" 

       index, point = self._side.check_new_controlPoint_at (pos_x, pos_y)
       
       if index is not None: 

            self._side.add_controlPoint (index, point)
            # _finished will do the rest - and init complete refresh
            self._finished_point()
 

    def _delete_point (self, aPoint : Movable_Point):
        """ slot - point is should be deleted """
        # overloaded - don't delete point 1 
        if aPoint.id == 1: return
        super()._delete_point (aPoint)    

        px, py = self.jpoints_xy()
        self._side.set_controlPoints (px, py)   

        # _finished will do the rest - and init complete refresh
        self._finished_point()


    def _finished_point (self, aPoint = None):
        """ slot - point move is finished """

        # overloaded - update airfoil geometry 
        self._airfoil.geo.finished_change_of (self._side)      

        super()._finished_point(aPoint)



# -------- concrete sub classes ------------------------


class Airfoil_Artist (Artist):
    """Plot the airfoils contour  """


    def __init__ (self, *args, 
                  show_points = False,
                  **kwargs):

        self._show_panels = False                       # show ony panels 
        self._show_points = show_points is True         # show coordinate points

        super().__init__ (*args, **kwargs)


    @property
    def show_panels(self): return self._show_panels
    def set_show_panels (self, aBool): 
        self._show_panels = aBool 
        if self._show_panels: 
            self.set_show_points (False)


    @property
    def show_points (self): return self._show_points
    def set_show_points (self, aBool):
        """ user switch to show point (marker )
        """
        self._show_points = aBool is True 

        p : pg.PlotDataItem
        for p in self._plots:
            if isinstance (p, pg.PlotDataItem):
                if self.show_points:
                    p.setSymbol('o')
                else: 
                    p.setSymbol(None)
        self.plot()             # do refresh will show leading edge of spline 



    def set_current (self, aLineLabel):
        # tries to set a highlighted airfoil to section with name ''aLineLabel' 
        if (not aLineLabel is None and aLineLabel != self._curLineLabel):    # only when changed do something
            self._curLineLabel = aLineLabel
            if self.show:                       # view is switched on by user? 
                self.plot ()


    @property
    def airfoils (self) -> list [Airfoil]: return self.data_list


    def _plot (self): 
    
        color_palette = random_colors (len(self.airfoils))

        # are there many airfoils - one of them is DESIGN? 

        airfoils_with_design = False 
        for airfoil in self.airfoils:
            if airfoil.usedAsDesign:
                airfoils_with_design = True 
                break

        for iair, airfoil in enumerate (self.airfoils):
            if (airfoil.isLoaded):

                # no legend if it is only one airfoil 

                if len(self.airfoils) == 1:
                    label = None                                    # suppress legend 
                else: 
                    label = _label_airfoil (self.airfoils, airfoil)

                # set color, width, ZValue, symbol style depending on airfoil usage and no of airfoils  

                color = _color_airfoil (self.airfoils, airfoil)
                if color is None: color = color_palette [iair]

                # default 
                width = 1
                antialias = False
                zValue = 1

                if airfoil.usedAs == usedAs.FINAL:
                    zValue = 5                                      # final top most 
                elif airfoils_with_design:
                    if airfoil.usedAsDesign:
                        width = 2
                        antialias = True
                        zValue = 3
                    elif airfoil.usedAs != usedAs.NORMAL:
                        color = QColor (color).darker (110)
                elif airfoil.usedAs == usedAs.NORMAL:
                    width = 2
                    antialias = True

                pen = pg.mkPen(color, width=width)

                sPen, sBrush, sSize = pg.mkPen(color, width=1), 'black', 7
                s = 'o' if self.show_points else None 

                # plot contour and fill airfoil if it's only one 
                #   use geometry.xy to refelect changes in diesign airfoil

                if len(self.airfoils) == 1: 

                    # if there is only one airfoil, fill the airfoil contour with a soft color tone  
                    brush = pg.mkBrush (color.darker (600))
                    self._plot_dataItem  (airfoil.geo.x, airfoil.geo.y, name=label, pen = pen, 
                                          symbol=s, symbolSize=sSize, symbolPen=sPen, symbolBrush=sBrush, 
                                          fillLevel=0.0, fillBrush=brush, antialias = antialias,
                                          zValue=zValue)
                    
                    # plot note if reflexed or rearloaded
                    self._plot_reflexed_rearloaded (airfoil, color)

                else: 
                    self._plot_dataItem  (airfoil.geo.x, airfoil.geo.y, name=label, pen = pen, 
                                          symbol=s, symbolSize=sSize, symbolPen=sPen, symbolBrush=sBrush,
                                          antialias = antialias,
                                          zValue=zValue)

                # optional plot of real LE defined by spline 

                if self.show_points and airfoil.geo.isSplined:
                    if airfoil.isNormalized:
                        brushcolor = "limegreen"
                        text = None
                    else: 
                        brushcolor = "yellow"
                        text="LE spline"
                    self._plot_point (airfoil.geo.le_real, color=color, brush=brushcolor,
                                      text=text,anchor=(0.5,1) )


    def _plot_reflexed_rearloaded (self, airfoil : Airfoil, color : QColor): 
        """ plot note if reflexed or rearloaded"""

        textColor = color.darker (140)                  #Artist.COLOR_LEGEND

        if airfoil.isReflexed: 
            x = 0.8
            y = airfoil.geo.upper.yFn (x) 
            self._plot_point ((x,y), size=0, text="Reflexed", anchor=(0.5,2.0), textColor=textColor) 

        elif airfoil.isRearLoaded: 
            x = 0.8
            y = airfoil.geo.lower.yFn (x) 
            self._plot_point ((x,y), size=0, text="Rearloaded", anchor=(0.5,-1.0), textColor=textColor)



class Flap_Artist (Artist):
    """Plot the flapped airfoil based on Flapper data  """

    def __init__ (self, *args, **kwargs):
        super().__init__ (*args, **kwargs)


    @property
    def airfoils (self) -> list [Airfoil]: return self.data_list

    @property
    def design_airfoil (self) -> Airfoil:
        for airfoil in self.airfoils:
            if airfoil.usedAsDesign:
                return airfoil 
    
    @property
    def flap_setter (self) -> Flap_Setter:
        return self.design_airfoil.flap_setter if self.design_airfoil else None 


    def _plot (self): 
    
        if self.flap_setter is None: return 

        flapped_airfoil = self.flap_setter.airfoil_flapped
        color = _color_airfoil ([], self.design_airfoil)

        if flapped_airfoil is not None:

            # plot flapped airfoil 

            pen   = pg.mkPen(color, width=1, style=Qt.PenStyle.DashLine)
            label = f"{self.design_airfoil.name_to_show} flapped"

            self._plot_dataItem  (flapped_airfoil.x, flapped_airfoil.y, name=label, pen = pen, 
                                    antialias = False, zValue=5)

            # plot flap angle 

            x = (1.0 + flapped_airfoil.x[0]) / 2
            y = flapped_airfoil.y[0] / 2

            self._plot_point ((x,y), size=0,text=f"{self.flap_setter.flap_angle:.1f}°", anchor=(-0.1, 0.5))

        # plot hinge point at the initial, unflapped airfoil

        x = self.flap_setter.x_flap
        airfoil_base = self.flap_setter.airfoil_base

        y_base = airfoil_base.geo.lower.yFn(x)
        thick  = airfoil_base.geo.thickness.yFn(x)
        y = y_base + self.flap_setter.y_flap * thick 

        self._plot_point ((x,y), color=color, size=10,text=f"Hinge {self.flap_setter.x_flap:.1%}" )



class Bezier_Artist (Artist):
    """Plot and edit airfoils Bezier control points """

    sig_bezier_changed     = pyqtSignal()           # bezier curve changed 


    @property
    def airfoils (self) -> list [Airfoil]: return self.data_list

    def refresh_from_side (self, aLinetype):

        p : Movable_Side_Bezier
        for p in self._plots: 
            if p._side.type == aLinetype:
                p.refresh()

    def _plot (self): 
    
        airfoil: Airfoil

        for airfoil in self.airfoils:
            if airfoil.isBezierBased and airfoil.isLoaded:

                color = _color_airfoil (self.airfoils, airfoil)
                movable = airfoil.usedAsDesign and self.show_mouse_helper

                side : Side_Airfoil_Bezier
                for side in [airfoil.geo.lower, airfoil.geo.upper]:     # paint upper on top 

                    p = Movable_Side_Bezier (airfoil, side, color=color, movable=movable,
                                              on_changed=self.sig_bezier_changed.emit) 
                    self._add(p)
 
                    # connect to mouse click in scene to add a new Bezier control point 

                    if movable:
                        sc : pg.GraphicsScene = p.scene()
                        sc.sigMouseClicked.connect (p.scene_clicked)



class Bezier_Deviation_Artist (Artist):
    """Plot deviation of Bezier curves to target airfoil """

    @property
    def airfoils (self) -> list [Airfoil]: return self.data_list


    def _plot (self): 
    
        from airfoil_dialogs    import Matcher

        # check - is there a design airfoil and a target (normal) airfoil 

        geo_from   : Geometry = None 
        geo_target : Geometry = None 

        for airfoil in self.airfoils:
            if airfoil.usedAsDesign:
                geo_from = airfoil.geo 
            elif airfoil.usedAs == usedAs.NORMAL:
                geo_target = airfoil.geo

        if not geo_from or not geo_target: return 

        # plot difference as thick lines at deviation check coordinates 

        label = "Deviation x 10"
        color = COLOR_WARNING  
        color.setAlphaF (0.6)

        for side in [Line.Type.UPPER, Line.Type.LOWER]:     

            if side == Line.Type.UPPER:
                from_line   = geo_from.upper  
                target_line = geo_target.upper
            else:
                from_line   = geo_from.lower  
                target_line = geo_target.lower

            # get deviation from Matcher 

            x, y, devi = Line.deviation_to (from_line, target_line, reduce_points=False, fast=True)

            # build array of coordinate pairs for fast plot -> connect='pairs'

            x_dbl = np.zeros ((len(x)-2)*2)
            y_dbl = np.zeros ((len(x)-2)*2)

            j = 0 
            for i in range (1, len(x)-1):
                if abs(devi[i]) > 1e-6:                     # avoid pyqtgraph artefacts
                    x_dbl[j] = x[i]
                    y_dbl[j] = y[i]
                    j += 1
                    x_dbl[j] = x[i]
                    y_dbl[j] = y[i] - (devi[i] * 10) 
                    j += 1

            x_dbl = x_dbl [:j]
            y_dbl = y_dbl [:j]

            if len(x_dbl):
                self._plot_dataItem  (x_dbl, y_dbl, pen=pg.mkPen(color, width=3), name=label, 
                                        antialias=False, zValue=1, connect='pairs')    




class Hicks_Henne_Artist (Artist):
    """Plot and edit airfoils Hicks Henne functions of airfoils  """

    @property
    def airfoils (self) -> list [Airfoil]: return self.data_list

    def _plot (self): 
    
        y_factor = 20 

        for airfoil in self.airfoils:
            if airfoil.isHicksHenneBased and airfoil.isLoaded:

                color_airfoil = _color_airfoil (self.airfoils, airfoil)

                side : Side_Airfoil_HicksHenne
                for side in [airfoil.geo.upper, airfoil.geo.lower]:     # paint upper on top 

                    hh : HicksHenne
                    for ih, hh in enumerate(side.hhs):

                        color = color_in_series (color_airfoil,ih, len(side.hhs), delta_hue=0.2)    
                        style = Qt.PenStyle.DashDotDotLine if side.isUpper else Qt.PenStyle.DashLine
                        pen   = pg.mkPen(color, width=0.7, style=style)

                        if ih == 0:
                            label = f"Hicks Henne upper x {y_factor}" if side.isUpper else f"Hicks Henne lower x {y_factor}"
                        else: 
                            label = None 

                        # plot hh function 
                        x = side.x 
                        y = hh.eval (x) * y_factor # + delta_y

                        self._plot_dataItem  (x, y,pen = pen, antialias = False, zValue=4, name=label)

                        # plot maximum marker 

                        x = hh.location
                        y = hh.strength  * y_factor   

                        text = f"HH{ih+1}"
                        anchor = (-0.05,1.05) if side.isUpper else (-0.05,-0.05)

                        self._plot_point ((x, y), symbol = '+', color=color, text=text, textColor=color, anchor=anchor,
                                          zValue=4)

 



class Curvature_Artist (Artist):
    """
    Plot curvature (top or bottom) of an airfoil
    """
    name = 'Curvature' 

    def __init__ (self, *args, show_derivative=False, **kwargs):

        self._show_upper = True                     # show upper side 
        self._show_lower = True                     # show lower side 
        self._show_derivative = show_derivative     # show derivative of curvature 
 
        super().__init__ (*args, **kwargs)

    @property
    def show_upper(self): return self._show_upper
    def set_show_upper (self, aBool): 
        self._show_upper = aBool
        self.plot() 

    @property
    def show_lower(self): return self._show_lower
    def set_show_lower (self, aBool): 
        self._show_lower = aBool 
        self.plot()

    @property
    def show_derivative(self): return self._show_derivative
    def set_show_derivative (self, aBool): 
        self._show_derivative = aBool 
        self.plot()

    @property
    def airfoils (self) -> list [Airfoil]: return self.data_list


    def _plot (self): 

        nairfoils = len(self.airfoils)
        
        airfoil: Airfoil

        for airfoil in self.airfoils:

            color = _color_airfoil (self.airfoils, airfoil)

            sides = []
            if self.show_upper: sides.append (airfoil.geo.curvature.upper)
            if self.show_lower: sides.append (airfoil.geo.curvature.lower)

            side : Line
            for side in sides:
                x = side.x
                y = side.y      
                if side.type == Line.Type.UPPER:
                    pen = pg.mkPen(color, width=1, style=Qt.PenStyle.SolidLine)
                else: 
                    pen = pg.mkPen(color, width=1, style=Qt.PenStyle.DashLine)

                label = f"{side.name} - {airfoil.name_to_show}"
                self._plot_dataItem (x, y, name=label, pen=pen)

                # plot derivative1 of curvature ('spikes') 

                if self.show_derivative and (nairfoils == 1 or airfoil.usedAsDesign):
                    pen = QPen (pen)
                    pen.setColor (QColor('red'))
                    name = f"{side.name} - Derivative"
                    self._plot_dataItem (x, -derivative1(x,y), name=name, pen=pen)

                # plot max points at le and te and reversals

                self._plot_le_te_max_point (side, color )
                self._plot_reversals (side, color)



    def _plot_le_te_max_point (self, aSide : Line, color ):
        """ plot the max values at LE and te"""

        point_color = QColor (color).darker (150)

        # le 
        if aSide.isUpper:
            anchor = (0,1)
        else: 
            anchor = (0,0)
        text = f"max {aSide.name} {aSide.max_xy[1]:.0f}"
        self._plot_point (aSide.max_xy, color=point_color, text=text, anchor=anchor,
                          textColor=Artist.COLOR_LEGEND)

        # te 
        if aSide.isUpper:
            anchor = (1,1)
        else: 
            anchor = (1,0)
        text = f"max {aSide.name} {aSide.te[1]:.1f}"
        self._plot_point (aSide.te, color=point_color, text=text, anchor=anchor,
                          textColor=Artist.COLOR_LEGEND)


    def _plot_reversals (self, side : Line, color):
        """ annotate reversals of curvature """

        point_color = QColor (color).darker (150)

        reversals = side.reversals()
        if reversals:
            for reversal_xy in reversals: 
                if side.isUpper:
                    anchor = (1,1.2)
                else: 
                    anchor = (1,-0.2)
                self._plot_point (reversal_xy, color=point_color, size=2, text="R", anchor=anchor,
                                  textColor=Artist.COLOR_LEGEND)


# class Difference_Artist (Airfoil_Line_Artist):
#     """Plot the y-difference of two airfoils 

#         2nd airfoil is Bezier based airfoil 
#         1st is reference or original airfoil from where x-stations are taken  
#     """

#     @property
#     def airfoil (self) -> Airfoil_Bezier: 
#         return self.airfoils[1] 
    
#     @property
#     def ref_airfoil (self) -> Airfoil : 
#         return self.airfoils[0] 
    

#     def _get_difference (self, side_ref: Side_Airfoil, side_actual: Side_Airfoil_Bezier):
#         # calculate difference at y-stations of reference airfoil 
#         diff  = np.zeros (len(side_ref.x))
#         for i, x in enumerate(side_ref.x):
#             diff [i] = side_actual.bezier.eval_y_on_x (x, fast=True) - side_ref.y[i]
#         return diff 


#     def _plot (self): 

#         if len(self.airfoils) != 2 : return 

#         self.set_showLegend (False)                             # no legend 
#         color = _color_airfoil (self.airfoil.usedAs)
#         linewidth=0.8

#         if self.upper:
#             x = self.ref_airfoil.geo.upper.x
#             y = 10 * self._get_difference (self.ref_airfoil.geo.upper, self.airfoil.geo.upper )
#             p = self.ax.plot (x, y, ls_difference, color = color, 
#                             linewidth= linewidth, **self._marker_style)
#             self._add(p)

#         if self.lower:
#             x = self.ref_airfoil.geo.lower.x
#             y = 10 * self._get_difference (self.ref_airfoil.geo.lower, self.airfoil.geo.lower ) 
#             p = self.ax.plot (x, y, ls_difference, color = color, 
#                             linewidth= linewidth, **self._marker_style)
#             self._add(p)




class Airfoil_Line_Artist (Artist, QObject):
    """
    Plot thickness, camber line of an airfoil, print max values 
    """

    sig_geometry_changed     = pyqtSignal()          # airfoil data changed 

    @property
    def airfoils (self) -> list [Airfoil]: return self.data_list


    def _plot (self): 

        airfoil: Airfoil

        for iair, airfoil in enumerate (self.airfoils):

            color = _color_airfoil (self.airfoils, airfoil)
            color.setAlphaF (0.8)

            # plot all 'lines' of airfoil 

            for line in airfoil.geo.lines_dict.values():
                
                style = _linestyle_of (line._type)
                if airfoil.usedAsDesign:
                    zValue = 5                                  # plot design on top 
                else:
                    zValue = 4 

                is_upper = line.type == Line.Type.UPPER
                is_lower = line.type == Line.Type.LOWER
                is_upper_lower = is_upper or is_lower

                if iair == 0 and not is_upper_lower:            # line legend only for the first airfoil 
                    name = line.name
                else: 
                    name = None 

                # plot upper and lower only for design (to visualize move highpoint)

                if airfoil.usedAsDesign or not is_upper_lower:
                    pen = pg.mkPen(color, width=1, style=style)
                    p = self._plot_dataItem (line.x, line.y, pen = pen, name = name, zValue=zValue)
                else: 
                    p = None

                # plot its highpoint 

                ph = Movable_Highpoint (airfoil.geo, line, p, 
                                            movable=airfoil.usedAsDesign, color=color,
                                            on_changed=self.sig_geometry_changed.emit )
                self._add (ph) 


                # te gap point for DESIGN 

                if airfoil.usedAsDesign and is_upper:
                    pt = Movable_TE_Point (airfoil.geo, p, 
                                            movable=airfoil.usedAsDesign, color=color,
                                            on_changed=self.sig_geometry_changed.emit )
                    self._add (pt) 


            # plot le circle 

            radius = airfoil.geo.le_radius
            
            circle_item = self._plot_circle (radius, 0, color=color, size=2*radius, 
                                            style=Qt.PenStyle.DotLine)
            pl = Movable_LE_Point (airfoil.geo, circle_item, 
                                    movable=airfoil.usedAsDesign, color=color,
                                    on_changed=self.sig_geometry_changed.emit )
            self._add(pl) 



class Polar_Artist (Artist):
    """Plot the polars of airfoils """


    def __init__ (self, axes, modelFn, 
                  xyVars = (var.CD, var.CL), 
                  **kwargs):
        super().__init__ (axes, modelFn, **kwargs)

        self._show_points = False                       # show point marker 
        self._xyVars = xyVars                           # definition of x,y axis


    @property
    def show_points(self): return self._show_points
    def set_show_points (self, aBool): 
        self._show_points = aBool 
        self.refresh()

    @property
    def xyVars(self): return self._xyVars
    def set_xyVars (self, xyVars: Tuple[var, var]): 
        """ set new x, y variables for polar """
        self._xyVars = xyVars 
        self.refresh()


    @property
    def airfoils (self) -> list [Airfoil]: return self.data_list


    def _plot (self): 
        """ do plot of airfoil polars in the prepared axes  """

        if not self.airfoils : return 

        # cancel all polar generations which are not for this current set of airfoils 
        #   to avoid to many os worker threads 

        Polar_Task.terminate_instances_except_for (self.airfoils)

        # load or generate polars which are not loaded up to now

        for airfoil in self.airfoils: 
            polarSet = airfoil.polarSet
            if polarSet:
                polarSet.load_or_generate_polars ()
            else:
                logger.debug (f"{airfoil} has no polarSet to plot")

        # plot polars of airfoils

        nPolar_plotted      = 0 
        nPolar_generating = 0                     # is there a polar in calculation 
        error_msg           = []  

        airfoil: Airfoil
        for airfoil in self.airfoils:

            if airfoil.polarSet:
 
                color_airfoil = _color_airfoil (self.airfoils, airfoil)

                # first filter only visible polars to get number of polars and plot 

                polarSet : Polar_Set = airfoil.polarSet
                polars = polarSet.polars

                for iPolar, polar in enumerate(polars [::-1]): 

                    if not polar.isLoaded: 
                        nPolar_generating += 1
                    elif polar.error_occurred:
                        # in error_msg could be e.g. '<' 
                        error_msg.append (f"'{airfoil.name_to_show} - {polar.name}': {html.escape(polar.error_reason)}")
                    else: 

                        # generate increasing color hue value for the polars of an airfoil 
                        color = color_in_series (color_airfoil, iPolar, len(polars), delta_hue=0.1)

                        self._plot_polar (self.airfoils, airfoil, polar, color)

                        nPolar_plotted += 1

        # show error messages 

        if error_msg:
            text = '<br>'.join (error_msg)          
            self._plot_text (text, color=COLOR_ERROR, itemPos=(0.5,0.5))

        # show generating message 

        if nPolar_generating > 0: 
            if nPolar_generating == 1:
                text = f"Generating polar"
            else: 
                text = f"Generating {nPolar_generating} polars"
            self._plot_text (text, color= "dimgray", fontSize=self.SIZE_HEADER, itemPos=(0.5, 1))

            # refresh for new polars 
            # QTimer().singleShot(1000, self.check_for_new_polars)     # delayed emit 



    def _plot_polar (self, airfoils: list[Airfoil], airfoil : Airfoil, polar: Polar, color): 
        """ plot a single polar"""

        airfoils_with_design = False 
        for a in airfoils:
            if a.usedAsDesign: airfoils_with_design = True 

        # build nice label 

        label = f"{_label_airfoil (airfoils, airfoil)} - {polar.name}" 

        if not polar.isLoaded:
            label = label + ' generating'                       # async polar generation  

        # set linewidth 

        antialias = True

        if self._show_points:
            linewidth=0.5
        elif airfoil.usedAs == usedAs.FINAL:  
            linewidth=1.5
            antialias = True
        elif airfoil.usedAs == usedAs.DESIGN:  
            linewidth=1.5
            antialias = True
        elif airfoil.usedAs == usedAs.NORMAL and not airfoils_with_design:  
            linewidth=1.5
            antialias = True
        else:
            linewidth=1.0

        # NORMAl and DESIGN polar above other polars 

        if airfoil.usedAs == usedAs.FINAL:
            zValue = 5
        elif airfoil.usedAs == usedAs.DESIGN:
            zValue = 3
        elif airfoil.usedAs == usedAs.NORMAL:
            zValue = 2
        else: 
            zValue = 1

        # finally plot 

        pen = pg.mkPen(color, width=linewidth)
        sPen, sBrush, sSize = pg.mkPen(color, width=1), 'black', 7
        s = 'o' if self.show_points else None 

        x,y = polar.ofVars (self.xyVars)

        self._plot_dataItem  (x, y, name=label, pen = pen, 
                                symbol=s, symbolSize=sSize, symbolPen=sPen, symbolBrush=sBrush,
                                antialias = antialias, zValue=zValue)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

    Spline pytest classes

"""

import pytest

import numpy as np 
from spline import *

class Test_Spline:

    def test_spline_1D (self): 

        x = [  0, 0.5,  2,  3,  4,  5,  7]  
        y = [  0,  3,  0,  2,  0,  2,  0]  

        # scalar 
        spl = Spline1D (x,y, boundary="natural")
        xnew = 1.0
        ynew = spl.eval(xnew, der=0) 
        assert ynew == 2.9546299523643555
        print ("1D natural ", xnew, ynew)


        # scalar 
        spl = Spline1D (x,y, boundary="notaknot")
        xnew = 1.0
        ynew = spl.eval(xnew, der=0) 
        assert ynew == 2.642301710730949
        print ("1D notaknot", xnew, ynew)

        # array
        xnew = np.linspace( x[0], x[-1] , 10)
        ynew = spl.eval(xnew, der=0) 
        print ("1D notaknot der=0", np.sum (ynew), "   x sum", np.sum(xnew))
        assert np.sum (ynew) == 15.463452566096425

        ynew = spl.eval(xnew, der=1) 
        print ("1D notaknot der=1", np.sum (ynew))
        assert np.sum (ynew) == 0.07395887333679596

        ynew = spl.eval(xnew, der=2) 
        print ("1D notaknot der=2", np.sum (ynew))
        assert np.sum (ynew) == -49.18195956454122

        c  = spl.curvature(xnew) 
        print ("1D notaknot curv", np.sum (c))
        assert np.sum (c) == -6.761005232410417

        pass



    def test_spline_2D (self): 

        x = [  0, 0.5,  2,  3,  4,  5,  7]  
        y = [  0,  3,  0,  2,  0,  2,  0]  

        # array 
        spl = Spline2D (x,y, boundary="notaknot")

        u = np.linspace( 0, 1 , 10)

        xnew, ynew = spl.eval (u)
        assert np.sum (xnew) == 28.86562563555188
        assert np.sum (ynew) == 12.849386740364192
        # print ("2D notaknot", np.sum(xnew), np.sum(ynew))

        curv = spl.curvature (u)
        assert np.sum(curv) == -10.268879474628818
        # print ("2D notaknot curvature", np.sum(curv), curv)

        # xnew = np.linspace( x[0], x[-1] , 10)
        # ynew = spl.eval(xnew, der=0) 
        # print ("notaknot der=0", np.sum (ynew), "   x sum", np.sum(xnew))
        # assert np.sum (ynew) == 15.463452566096425

        # ynew = spl.eval(xnew, der=1) 
        # print ("notaknot der=1", np.sum (ynew))
        # assert np.sum (ynew) == 0.07395887333679596

        # ynew = spl.eval(xnew, der=2) 
        # print ("notaknot der=2", np.sum (ynew))
        # assert np.sum (ynew) == -49.18195956454122

        # c  = spl.curvature(xnew) 
        # print ("notaknot curv", np.sum (c))
        # assert np.sum (c) == -6.761005232410417

        pass


# Main program for testing 
if __name__ == "__main__":

    test = Test_Spline()
    test.test_spline_1D()
    test.test_spline_2D()
